# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
import threading
import time
import uuid
from typing import Optional

from libcloud.compute.drivers.gce import GCEFailedNode
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from toil.jobStores.googleJobStore import GoogleJobStore
from toil.lib.compatibility import compat_bytes_recursive
from toil.lib.conversions import human2bytes
from toil.lib.web import web_session
from toil.provisioners import NoSuchClusterException
from toil.provisioners.abstractProvisioner import AbstractProvisioner, Shape
from toil.provisioners.node import Node

logger = logging.getLogger(__name__)
logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)


class GCEProvisioner(AbstractProvisioner):
    """
    Implements a Google Compute Engine Provisioner using libcloud.
    """

    NODE_BOTO_PATH = "/root/.boto"  # boto file path on instances
    SOURCE_IMAGE = b"projects/kinvolk-public/global/images/family/flatcar-stable"

    def __init__(
        self,
        clusterName,
        clusterType,
        zone,
        nodeStorage,
        nodeStorageOverrides,
        sseKey,
        enable_fuse,
    ):
        self.cloud = "gce"
        self._sseKey = sseKey

        # Call base class constructor, which will call createClusterSettings()
        # or readClusterSettings()
        super().__init__(
            clusterName,
            clusterType,
            zone,
            nodeStorage,
            nodeStorageOverrides,
            enable_fuse,
        )

    def supportedClusterTypes(self):
        return {"mesos"}

    def createClusterSettings(self):
        # All we need to do is read the Google credentials we need to provision
        # things
        self._readCredentials()

    def readClusterSettings(self):
        """
        Read the cluster settings from the instance, which should be the leader.
        See https://cloud.google.com/compute/docs/storing-retrieving-metadata for details about
        reading the metadata.
        """
        metadata_server = "http://metadata/computeMetadata/v1/instance/"
        metadata_flavor = {"Metadata-Flavor": "Google"}
        zone = web_session.get(metadata_server + "zone", headers=metadata_flavor).text
        self._zone = zone.split("/")[-1]

        project_metadata_server = "http://metadata/computeMetadata/v1/project/"
        self._projectId = web_session.get(
            project_metadata_server + "project-id", headers=metadata_flavor
        ).text

        # From a GCE instance, these values can be blank. Only the projectId is needed
        self._googleJson = ""
        self._clientEmail = ""

        self._tags = web_session.get(
            metadata_server + "description", headers=metadata_flavor
        ).text
        tags = json.loads(self._tags)
        self.clusterName = tags["clusterName"]
        self._gceDriver = self._getDriver()
        self._instanceGroup = self._gceDriver.ex_get_instancegroup(
            self.clusterName, zone=self._zone
        )

        leader = self.getLeader()
        self._leaderPrivateIP = leader.privateIP

        # The location of the Google credentials file on instances.
        self._credentialsPath = GoogleJobStore.nodeServiceAccountJson
        self._keyName = "core"  # key name leader users to communicate with works
        self._botoPath = (
            self.NODE_BOTO_PATH
        )  # boto credentials (used if reading an AWS bucket)

        # Let the base provisioner work out how to deploy duly authorized
        # workers for this leader.
        self._setLeaderWorkerAuthentication()

    def _readCredentials(self):
        """
        Get the credentials from the file specified by GOOGLE_APPLICATION_CREDENTIALS.
        """
        self._googleJson = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
        if not self._googleJson:
            raise RuntimeError("GOOGLE_APPLICATION_CREDENTIALS not set.")
        try:
            with open(self._googleJson) as jsonFile:
                self.googleConnectionParams = json.loads(jsonFile.read())
        except:
            raise RuntimeError(
                "GCEProvisioner: Could not parse the Google service account json file %s"
                % self._googleJson
            )

        self._projectId = self.googleConnectionParams["project_id"]
        self._clientEmail = self.googleConnectionParams["client_email"]
        self._credentialsPath = self._googleJson
        self._clearLeaderWorkerAuthentication()  # TODO: Why are we doing this?
        self._gceDriver = self._getDriver()

    def _write_file_to_cloud(self, key: str, contents: bytes) -> str:
        raise NotImplementedError(
            "The gceProvisioner doesn't support _write_file_to_cloud()."
        )

    def _get_user_data_limit(self) -> int:
        # See: https://cloud.google.com/compute/docs/metadata/setting-custom-metadata#limitations
        return human2bytes("256KB")

    def launchCluster(self, leaderNodeType, leaderStorage, owner, **kwargs):
        """
        In addition to the parameters inherited from the abstractProvisioner,
        the Google launchCluster takes the following parameters:
        keyName: The key used to communicate with instances
        botoPath: Boto credentials for reading an AWS jobStore (optional).
        network: a network (optional)
        vpcSubnet: A subnet (optional).
        use_private_ip: even though a public ip exists, ignore it (optional)
        """
        if "keyName" not in kwargs:
            raise RuntimeError("A keyPairName is required for the GCE provisioner.")
        self._keyName = kwargs["keyName"]
        if "botoPath" in kwargs:
            self._botoPath = kwargs["botoPath"]
        self._vpcSubnet = kwargs.get("vpcSubnet", None)
        self._network = kwargs.get("network", None)
        self._use_private_ip = kwargs.get("use_private_ip", None)

        # Throws an error if cluster exists
        self._instanceGroup = self._gceDriver.ex_create_instancegroup(
            self.clusterName, self._zone
        )
        logger.debug("Launching leader")

        # GCE doesn't have a dictionary tags field. The tags field is just a string list.
        # Therefore, dumping tags into the description.
        tags = {"Owner": self._keyName, "clusterName": self.clusterName}
        if "userTags" in kwargs:
            tags.update(kwargs["userTags"])
        self._tags = json.dumps(tags)

        metadata = {
            "items": [
                {"key": "user-data", "value": self._getIgnitionUserData("leader")}
            ]
        }
        imageType = "flatcar-stable"
        sa_scopes = [{"scopes": ["compute", "storage-full"]}]
        disk = {}
        disk["initializeParams"] = {
            "sourceImage": self.SOURCE_IMAGE,
            "diskSizeGb": leaderStorage,
        }
        disk.update({"boot": True, "autoDelete": True})
        name = "l" + str(uuid.uuid4())

        leader = self._gceDriver.create_node(
            name,
            leaderNodeType,
            imageType,
            location=self._zone,
            ex_service_accounts=sa_scopes,
            ex_metadata=compat_bytes_recursive(metadata),
            ex_network=self._network,
            ex_subnetwork=self._vpcSubnet,
            ex_disks_gce_struct=[compat_bytes_recursive(disk)],
            description=self._tags,
            ex_preemptible=False,
        )

        self._instanceGroup.add_instances([leader])
        self._leaderPrivateIP = leader.private_ips[0]  # needed if adding workers
        # self.subnetID = leader.subnet_id  # TODO: get subnetID

        # Wait for the appliance to start and inject credentials.
        leaderNode = Node(
            publicIP=leader.public_ips[0],
            privateIP=leader.private_ips[0],
            name=leader.name,
            launchTime=leader.created_at,
            nodeType=leader.size,
            preemptible=False,
            tags=self._tags,
            use_private_ip=self._use_private_ip,
        )
        leaderNode.waitForNode("toil_leader", keyName=self._keyName)
        leaderNode.copySshKeys(self._keyName)
        leaderNode.injectFile(
            self._credentialsPath, GoogleJobStore.nodeServiceAccountJson, "toil_leader"
        )
        if self._botoPath:
            leaderNode.injectFile(self._botoPath, self.NODE_BOTO_PATH, "toil_leader")
        # Download credentials
        self._setLeaderWorkerAuthentication(leaderNode)

        logger.debug("Launched leader")

    def getNodeShape(self, instance_type: str, preemptible=False) -> Shape:
        # TODO: read this value only once
        sizes = self._gceDriver.list_sizes(location=self._zone)
        sizes = [x for x in sizes if x.name == instance_type]
        assert len(sizes) == 1
        instanceType = sizes[0]

        disk = 0  # instanceType.disks * instanceType.disk_capacity * 2 ** 30
        if disk == 0:
            # This is an EBS-backed instance. We will use the root
            # volume, so add the amount of EBS storage requested forhe root volume
            disk = (
                self._nodeStorageOverrides.get(instance_type, self._nodeStorage) * 2**30
            )

        # Ram is in M.
        # Underestimate memory by 100M to prevent autoscaler from disagreeing with
        # mesos about whether a job can run on a particular node type
        memory = (instanceType.ram / 1000 - 0.1) * 2**30
        return Shape(
            wallTime=60 * 60,
            memory=memory,
            cores=instanceType.extra["guestCpus"],
            disk=disk,
            preemptible=preemptible,
        )

    @staticmethod
    def retryPredicate(e):
        """Not used by GCE"""
        return False

    def destroyCluster(self) -> None:
        """
        Try a few times to terminate all of the instances in the group.
        """
        logger.debug("Destroying cluster %s" % self.clusterName)
        instancesToTerminate = self._getNodesInCluster()
        attempts = 0
        while instancesToTerminate and attempts < 3:
            self._terminateInstances(instances=instancesToTerminate)
            instancesToTerminate = self._getNodesInCluster()
            attempts += 1

        # remove group
        instanceGroup = self._gceDriver.ex_get_instancegroup(
            self.clusterName, zone=self._zone
        )
        instanceGroup.destroy()

    def terminateNodes(self, nodes):
        if nodes:
            nodeNames = [n.name for n in nodes]
            instances = self._getNodesInCluster()
            instancesToKill = [i for i in instances if i.name in nodeNames]
            self._terminateInstances(instancesToKill)

    def addNodes(self, nodeTypes: set[str], numNodes, preemptible, spotBid=None) -> int:
        assert self._leaderPrivateIP

        # We don't support any balancing here so just pick one of the
        # equivalent node types
        node_type = next(iter(nodeTypes))

        # If keys are rsynced, then the mesos-slave needs to be started after the keys have been
        # transferred. The waitForKey.sh script loops on the new VM until it finds the keyPath file, then it starts the
        # mesos-slave. If there are multiple keys to be transferred, then the last one to be transferred must be
        # set to keyPath.
        keyPath = None
        botoExists = False
        if self._botoPath is not None and os.path.exists(self._botoPath):
            keyPath = self.NODE_BOTO_PATH
            botoExists = True
        elif self._sseKey:
            keyPath = self._sseKey

        if not preemptible:
            logger.debug("Launching %s non-preemptible nodes", numNodes)
        else:
            logger.debug("Launching %s preemptible nodes", numNodes)

        # kwargs["subnet_id"] = self.subnetID if self.subnetID else self._getClusterInstance(self.instanceMetaData).subnet_id
        userData = self._getIgnitionUserData("worker", keyPath, preemptible)
        metadata = {"items": [{"key": "user-data", "value": userData}]}
        imageType = "flatcar-stable"
        sa_scopes = [{"scopes": ["compute", "storage-full"]}]
        disk = {}
        disk["initializeParams"] = {
            "sourceImage": self.SOURCE_IMAGE,
            "diskSizeGb": self._nodeStorageOverrides.get(node_type, self._nodeStorage),
        }
        disk.update({"boot": True, "autoDelete": True})

        # TODO:
        #  - bug in gce.py for ex_create_multiple_nodes (erroneously, doesn't allow image and disk to specified)
        #  - ex_create_multiple_nodes is limited to 1000 nodes
        #    - use a different function
        #    - or write a loop over the rest of this function, with 1K nodes max on each iteration
        retries = 0
        workersCreated = 0
        # Try a few times to create the requested number of workers
        while numNodes - workersCreated > 0 and retries < 3:
            instancesLaunched = self.ex_create_multiple_nodes(
                "",
                node_type,
                imageType,
                numNodes - workersCreated,
                location=self._zone,
                ex_service_accounts=sa_scopes,
                ex_metadata=metadata,
                ex_disks_gce_struct=[disk],
                description=self._tags,
                ex_preemptible=preemptible,
            )
            failedWorkers = []
            for instance in instancesLaunched:
                if isinstance(instance, GCEFailedNode):
                    logger.error(
                        "Worker failed to launch with code %s. Error message: %s"
                        % (instance.code, instance.error)
                    )
                    continue

                node = Node(
                    publicIP=instance.public_ips[0],
                    privateIP=instance.private_ips[0],
                    name=instance.name,
                    launchTime=instance.created_at,
                    nodeType=instance.size,
                    preemptible=False,
                    tags=self._tags,
                    use_private_ip=self._use_private_ip,
                )  # FIXME: what should tags be set to?

                try:
                    self._injectWorkerFiles(node, botoExists)
                    logger.debug("Created worker %s" % node.publicIP)
                    self._instanceGroup.add_instances([instance])
                    workersCreated += 1
                except Exception as e:
                    logger.error(
                        f"Failed to configure worker {node.name}. Error message: {e}"
                    )
                    failedWorkers.append(instance)
            if failedWorkers:
                logger.error("Terminating %d failed workers" % len(failedWorkers))
                self._terminateInstances(failedWorkers)
            retries += 1

        logger.debug("Launched %d new instance(s)", numNodes)
        if numNodes != workersCreated:
            logger.error("Failed to launch %d worker(s)", numNodes - workersCreated)
        return workersCreated

    def getProvisionedWorkers(
        self, instance_type: Optional[str] = None, preemptible: Optional[bool] = None
    ):
        assert self._leaderPrivateIP
        entireCluster = self._getNodesInCluster(instance_type=instance_type)
        logger.debug("All nodes in cluster: %s", entireCluster)
        workerInstances = []
        for instance in entireCluster:
            if preemptible is not None:
                scheduling = instance.extra.get("scheduling")
                # If this field is not found in the extra meta-data, assume the node is not preemptible.
                if scheduling and scheduling.get("preemptible", False) != preemptible:
                    continue
            isWorker = True
            for ip in instance.private_ips:
                if ip == self._leaderPrivateIP:
                    isWorker = False
                    break  # don't include the leader
            if isWorker and instance.state == "running":
                workerInstances.append(instance)

        logger.debug("All workers found in cluster: %s", workerInstances)
        return [
            Node(
                publicIP=i.public_ips[0],
                privateIP=i.private_ips[0],
                name=i.name,
                launchTime=i.created_at,
                nodeType=i.size,
                preemptible=i.extra.get("scheduling", {}).get("preemptible", False),
                tags=None,
                use_private_ip=self._use_private_ip,
            )
            for i in workerInstances
        ]

    def getLeader(self):
        instances = self._getNodesInCluster()
        instances.sort(key=lambda x: x.created_at)
        try:
            leader = instances[0]  # assume leader was launched first
        except IndexError:
            raise NoSuchClusterException(self.clusterName)
        return Node(
            publicIP=leader.public_ips[0],
            privateIP=leader.private_ips[0],
            name=leader.name,
            launchTime=leader.created_at,
            nodeType=leader.size,
            preemptible=False,
            tags=None,
            use_private_ip=self._use_private_ip,
        )

    def _injectWorkerFiles(self, node, botoExists):
        """
        Set up the credentials on the worker.
        """
        node.waitForNode("toil_worker", keyName=self._keyName)
        node.copySshKeys(self._keyName)
        node.injectFile(
            self._credentialsPath, GoogleJobStore.nodeServiceAccountJson, "toil_worker"
        )
        if self._sseKey:
            node.injectFile(self._sseKey, self._sseKey, "toil_worker")
        if botoExists:
            node.injectFile(self._botoPath, self.NODE_BOTO_PATH, "toil_worker")

    def _getNodesInCluster(self, instance_type: Optional[str] = None):
        instanceGroup = self._gceDriver.ex_get_instancegroup(
            self.clusterName, zone=self._zone
        )
        instances = instanceGroup.list_instances()
        if instance_type:
            instances = [
                instance for instance in instances if instance.size == instance_type
            ]
        return instances

    def _getDriver(self):
        """Connect to GCE"""
        driverCls = get_driver(Provider.GCE)
        return driverCls(
            self._clientEmail,
            self._googleJson,
            project=self._projectId,
            datacenter=self._zone,
        )

    def _terminateInstances(self, instances):
        def worker(driver, instance):
            logger.debug("Terminating instance: %s", instance.name)
            driver.destroy_node(instance)

        threads = []
        for instance in instances:
            t = threading.Thread(target=worker, args=(self._gceDriver, instance))
            threads.append(t)
            t.start()

        logger.debug("... Waiting for instance(s) to shut down...")
        for t in threads:
            t.join()

    # MONKEY PATCH - This function was copied form libcloud to fix a bug.
    DEFAULT_TASK_COMPLETION_TIMEOUT = 180

    def ex_create_multiple_nodes(
        self,
        base_name,
        size,
        image,
        number,
        location=None,
        ex_network="default",
        ex_subnetwork=None,
        ex_tags=None,
        ex_metadata=None,
        ignore_errors=True,
        use_existing_disk=True,
        poll_interval=2,
        external_ip="ephemeral",
        ex_disk_type="pd-standard",
        ex_disk_auto_delete=True,
        ex_service_accounts=None,
        timeout=DEFAULT_TASK_COMPLETION_TIMEOUT,
        description=None,
        ex_can_ip_forward=None,
        ex_disks_gce_struct=None,
        ex_nic_gce_struct=None,
        ex_on_host_maintenance=None,
        ex_automatic_restart=None,
        ex_image_family=None,
        ex_preemptible=None,
    ):
        """
        Monkey patch to gce.py in libcloud to allow disk and images to be specified.
        Also changed name to a uuid below.
        The prefix 'wp' identifies preemptible nodes and 'wn' non-preemptible nodes.
        """
        # if image and ex_disks_gce_struct:
        #    raise ValueError("Cannot specify both 'image' and "
        #                     "'ex_disks_gce_struct'.")

        driver = self._getDriver()
        if image and ex_image_family:
            raise ValueError("Cannot specify both 'image' and " "'ex_image_family'")

        location = location or driver.zone
        if not hasattr(location, "name"):
            location = driver.ex_get_zone(location)
        if not hasattr(size, "name"):
            size = driver.ex_get_size(size, location)
        if not hasattr(ex_network, "name"):
            ex_network = driver.ex_get_network(ex_network)
        if ex_subnetwork and not hasattr(ex_subnetwork, "name"):
            ex_subnetwork = driver.ex_get_subnetwork(
                ex_subnetwork, region=driver._get_region_from_zone(location)
            )
        if ex_image_family:
            image = driver.ex_get_image_from_family(ex_image_family)
        if image and not hasattr(image, "name"):
            image = driver.ex_get_image(image)
        if not hasattr(ex_disk_type, "name"):
            ex_disk_type = driver.ex_get_disktype(ex_disk_type, zone=location)

        node_attrs = {
            "size": size,
            "image": image,
            "location": location,
            "network": ex_network,
            "subnetwork": ex_subnetwork,
            "tags": ex_tags,
            "metadata": ex_metadata,
            "ignore_errors": ignore_errors,
            "use_existing_disk": use_existing_disk,
            "external_ip": external_ip,
            "ex_disk_type": ex_disk_type,
            "ex_disk_auto_delete": ex_disk_auto_delete,
            "ex_service_accounts": ex_service_accounts,
            "description": description,
            "ex_can_ip_forward": ex_can_ip_forward,
            "ex_disks_gce_struct": ex_disks_gce_struct,
            "ex_nic_gce_struct": ex_nic_gce_struct,
            "ex_on_host_maintenance": ex_on_host_maintenance,
            "ex_automatic_restart": ex_automatic_restart,
            "ex_preemptible": ex_preemptible,
        }
        # List for holding the status information for disk/node creation.
        status_list = []

        for i in range(number):
            name = "wp" if ex_preemptible else "wn"
            name += str(uuid.uuid4())  # '%s-%03d' % (base_name, i)
            status = {"name": name, "node_response": None, "node": None}
            status_list.append(status)

        start_time = time.time()
        complete = False
        while not complete:
            if time.time() - start_time >= timeout:
                raise Exception(
                    "Timeout (%s sec) while waiting for multiple " "instances"
                )
            complete = True
            time.sleep(poll_interval)
            for status in status_list:
                # Create the node or check status if already in progress.
                if not status["node"]:
                    if not status["node_response"]:
                        driver._multi_create_node(status, node_attrs)
                    else:
                        driver._multi_check_node(status, node_attrs)
                # If any of the nodes have not been created (or failed) we are
                # not done yet.
                if not status["node"]:
                    complete = False

        # Return list of nodes
        node_list = []
        for status in status_list:
            node_list.append(status["node"])
        return node_list



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging
import os

import connexion  # type: ignore
from configargparse import ArgumentParser

from toil.lib.aws import get_current_aws_region, running_on_ec2, running_on_ecs
from toil.server.wes.toil_backend import ToilBackend
from toil.server.wsgi_app import run_app
from toil.version import version

logger = logging.getLogger(__name__)


def parser_with_server_options() -> argparse.ArgumentParser:
    parser = ArgumentParser(description="Toil server mode.")

    parser.add_argument(
        "--debug", action="store_true", default=False, help="Enable debug mode."
    )
    parser.add_argument(
        "--bypass_celery",
        action="store_true",
        default=False,
        help="Skip sending workflows to Celery and just run them under the "
        "server. For testing.",
    )
    parser.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="The host interface that the Toil server binds on. (default: '127.0.0.1').",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=8080,
        help="The port that the Toil server listens on. (default: 8080).",
    )
    parser.add_argument(
        "--swagger_ui",
        action="store_true",
        default=False,
        help="If True, the swagger UI will be enabled and hosted on the "
        "`{api_base_path}/ui` endpoint. (default: False)",
    )
    # CORS
    parser.add_argument(
        "--cors",
        action="store_true",
        default=False,
        help="Enable Cross Origin Resource Sharing (CORS). This should only be turned on "
        "if the server is intended to be used by a website or domain.",
    )
    parser.add_argument(
        "--cors_origins",
        type=str,
        default="*",
        help="Ignored if --cors is False. This sets the allowed origins for CORS. "
        "For details about CORS and its security risks, see: "
        "https://w3id.org/ga4gh/product-approval-support/cors. (default: '*')",
    )
    # production only
    parser.add_argument(
        "-w",
        "--workers",
        dest="workers",
        type=int,
        default=2,
        help="Ignored if --debug is True. The number of worker processes launched by the "
        "WSGI server. (default: 2).",
    )

    parser.add_argument(
        "--work_dir",
        type=str,
        default=os.path.join(os.getcwd(), "workflows"),
        help="The directory where workflows should be stored. This directory should be "
        "empty or only contain previous workflows. (default: './workflows').",
    )
    parser.add_argument(
        "--state_store",
        type=str,
        default=None,
        help="The local path or S3 URL where workflow state metadata should be stored. "
        "(default: in --work_dir)",
    )
    parser.add_argument(
        "--opt",
        "-o",
        type=str,
        action="append",
        default=[],
        help="Specify the default parameters to be sent to the workflow engine for each "
        "run. Options taking arguments must use = syntax. Accepts multiple values.\n"
        "Example: '--opt=--logLevel=CRITICAL --opt=--workDir=/tmp'.",
    )
    parser.add_argument(
        "--dest_bucket_base",
        type=str,
        default=None,
        help="Direct CWL workflows to save output files to dynamically generated "
        "unique paths under the given URL. Supports AWS S3.",
    )
    parser.add_argument(
        "--wes_dialect",
        type=str,
        default="standard",
        choices=["standard", "agc"],
        help="Restrict WES responses to a dialect compatible with clients that do not fully "
        "implement the WES standard. (default: 'standard')",
    )

    parser.add_argument("--version", action="version", version=version)
    return parser


def create_app(args: argparse.Namespace) -> "connexion.FlaskApp":
    """
    Create a "connexion.FlaskApp" instance with Toil server configurations.
    """
    flask_app = connexion.FlaskApp(
        __name__, specification_dir="api_spec/", options={"swagger_ui": args.swagger_ui}
    )

    flask_app.app.config["JSON_SORT_KEYS"] = False

    if args.cors:
        # enable cross origin resource sharing
        from flask_cors import CORS

        CORS(flask_app.app, resources={r"/ga4gh/*": {"origins": args.cors_origins}})

    # add workflow execution service (WES) API endpoints
    backend = ToilBackend(
        work_dir=args.work_dir,
        state_store=args.state_store,
        options=args.opt,
        dest_bucket_base=args.dest_bucket_base,
        bypass_celery=args.bypass_celery,
        wes_dialect=args.wes_dialect,
    )

    flask_app.add_api(
        "workflow_execution_service.swagger.yaml",
        resolver=connexion.Resolver(backend.resolve_operation_id),
    )  # noqa

    # add custom endpoints
    if isinstance(backend, ToilBackend):
        # We extend the WES API to allow presenting log data
        base_url = "/toil/wes/v1"
        flask_app.app.add_url_rule(
            f"{base_url}/logs/<run_id>/stdout", view_func=backend.get_stdout
        )
        flask_app.app.add_url_rule(
            f"{base_url}/logs/<run_id>/stderr", view_func=backend.get_stderr
        )
        # To be a well-behaved AGC engine we can implement the default status check endpoint
        flask_app.app.add_url_rule("/engine/v1/status", view_func=backend.get_health)
        # And we can provide lost humans some information on what they are looking at
        flask_app.app.add_url_rule("/", view_func=backend.get_homepage)

    return flask_app


def start_server(args: argparse.Namespace) -> None:
    """Start a Toil server."""

    # Explain a bit about who and where we are
    logger.info("Toil WES server version %s starting...", version)
    if running_on_ecs():
        logger.info("Environment appears to be Amazon ECS")
    if running_on_ec2():
        logger.info("Environment appears to be Amazon EC2")
    aws_region = get_current_aws_region()
    if aws_region:
        logger.info("AWS region appears to be: %s", aws_region)

    flask_app = create_app(args)

    host = args.host
    port = args.port

    if args.debug:
        flask_app.run(host=host, port=port)
    else:
        # start a production WSGI server
        run_app(
            flask_app.app,
            options={
                "bind": f"{host}:{port}",
                "workers": args.workers,
            },
        )



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Optional

from configargparse import ArgParser, ArgumentParser

from toil.batchSystems.abstractBatchSystem import (
    AbstractBatchSystem,
    UpdatedBatchJobInfo,
)
from toil.batchSystems.cleanup_support import BatchSystemCleanupSupport
from toil.batchSystems.options import OptionSetter
from toil.batchSystems.registry import add_batch_system_factory
from toil.common import Toil, addOptions
from toil.job import JobDescription
from toil.test import ToilTest

logger = logging.getLogger(__name__)


class FakeBatchSystem(BatchSystemCleanupSupport):
    @classmethod
    def supportsAutoDeployment(cls) -> bool:
        pass

    def issueBatchJob(
        self,
        command: str,
        job_desc: JobDescription,
        job_environment: Optional[dict[str, str]] = None,
    ) -> int:
        pass

    def killBatchJobs(self, jobIDs: list[int]) -> None:
        pass

    def getIssuedBatchJobIDs(self) -> list[int]:
        pass

    def getRunningBatchJobIDs(self) -> dict[int, float]:
        pass

    def getUpdatedBatchJob(self, maxWait: int) -> Optional[UpdatedBatchJobInfo]:
        pass

    def shutdown(self) -> None:
        pass

    @classmethod
    def add_options(cls, parser: ArgumentParser) -> None:
        parser.add_argument("--fake_argument", default="exists")

    @classmethod
    def setOptions(cls, setOption: OptionSetter) -> None:
        setOption("fake_argument")


class BatchSystemPluginTest(ToilTest):
    def test_batchsystem_plugin_installable(self):
        """
        Test that installing a batch system plugin works.
        :return:
        """

        def fake_batch_system_factory() -> type[AbstractBatchSystem]:
            return FakeBatchSystem

        add_batch_system_factory("fake", fake_batch_system_factory)

        parser = ArgParser()
        addOptions(parser)

        options = parser.parse_args(["test-jobstore", "--clean=always"])

        # try to install a batchsystem plugin with some arguments
        # if the arguments exists, the values should also exist in the config
        with Toil(options) as toil:
            self.assertEqual(toil.config.fake_argument == "exists", True)



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp

if __name__ == "__main__":
    # A
    A = Job()
    A.addChild(Job())
    A.addFollowOn(Job())

    # Encapsulate A
    A = A.encapsulate()

    # B is a job which needs to run after A and its successors
    B = Job()

    # With encapsulation A and its successor subgraph appear to be a single job, hence:
    A.addChild(B)

    jobstore: str = mkdtemp("tutorial_encapsulations2")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "INFO"
    options.clean = "always"

    with Toil(options) as toil:
        print(toil.start(A))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable
import collections
import logging
import os
from pathlib import Path
import random
from typing import cast, Optional, Union, Any, NoReturn

import pytest

from toil.common import Toil
from toil.exceptions import FailedJobsException
from toil.job import (
    FunctionWrappingJob,
    Job,
    JobFunctionWrappingJob,
    JobGraphDeadlockException,
    ServiceHostJob,
    Promise,
)
from toil.lib.misc import FileDescriptorOrPath
from toil.test import pslow as slow

from pytest_subtests import SubTests


class TestJob:
    """Tests the job class."""

    @slow
    @pytest.mark.slow
    def testStatic(self, tmp_path: Path) -> None:
        r"""
        Create a DAG of jobs non-dynamically and run it. DAG is::

            A -> F
            \-------
            B -> D  \
             \       \
              ------- C -> E

        Follow on is marked by ``->``
        """
        outFile = tmp_path / "out"
        # Create the jobs
        A = Job.wrapFn(fn1Test, "A", outFile)
        B = Job.wrapFn(fn1Test, A.rv(), outFile)
        C = Job.wrapFn(fn1Test, B.rv(), outFile)
        D = Job.wrapFn(fn1Test, C.rv(), outFile)
        E = Job.wrapFn(fn1Test, D.rv(), outFile)
        F = Job.wrapFn(fn1Test, E.rv(), outFile)
        # Connect them into a workflow
        A.addChild(B)
        A.addChild(C)
        B.addChild(C)
        B.addFollowOn(E)
        C.addFollowOn(D)
        A.addFollowOn(F)

        # Create the runner for the workflow.
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.logLevel = "INFO"
        options.retryCount = 100
        options.badWorker = 0.5
        options.badWorkerFailInterval = 0.01
        # Run the workflow, the return value being the number of failed jobs
        Job.Runner.startToil(A, options)

        # Check output
        assert open(outFile).readline() == "ABCDEFG"

    def testStatic2(self, tmp_path: Path) -> None:
        r"""
        Create a DAG of jobs non-dynamically and run it. DAG is::

            A -> F
            \-------
            B -> D  \
             \       \
              ------- C -> E

        Follow on is marked by ``->``
        """
        outFile = tmp_path / "out"

        # Create the jobs
        A = Job.wrapFn(fn1Test, "A", outFile)
        B = Job.wrapFn(fn1Test, A.rv(), outFile)
        C = Job.wrapFn(fn1Test, B.rv(), outFile)
        D = Job.wrapFn(fn1Test, C.rv(), outFile)

        # Connect them into a workflow
        A.addChild(B)
        A.addFollowOn(C)
        C.addChild(D)

        # Create the runner for the workflow.
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.logLevel = "INFO"
        options.retryCount = 100
        options.badWorker = 0.5
        options.badWorkerFailInterval = 0.01
        # Run the workflow, the return value being the number of failed jobs
        Job.Runner.startToil(A, options)

        # Check output
        assert open(outFile).readline() == "ABCDE"

    @slow
    @pytest.mark.slow
    def testTrivialDAGConsistency(self, tmp_path: Path) -> None:
        options = Job.Runner.getDefaultOptions(tmp_path / "jobStore")
        options.clean = "always"
        options.logLevel = "debug"
        i = Job.wrapJobFn(trivialParent)
        with Toil(options) as toil:
            with pytest.raises(FailedJobsException):
                toil.start(i)

    @pytest.mark.timeout(300)
    def testDAGConsistency(self, tmp_path: Path) -> None:
        options = Job.Runner.getDefaultOptions(tmp_path / "jobStore")
        options.clean = "always"
        options.logLevel = "debug"
        i = Job.wrapJobFn(parent)
        with Toil(options) as toil:
            with pytest.raises(FailedJobsException):
                toil.start(i)

    @slow
    @pytest.mark.slow
    def testSiblingDAGConsistency(self, tmp_path: Path) -> None:
        """
        Slightly more complex case. The stranded job's predecessors are siblings instead of
        parent/child.
        """
        options = Job.Runner.getDefaultOptions(tmp_path / "jobStore")
        options.clean = "always"
        options.logLevel = "debug"
        i = Job.wrapJobFn(diamond)
        with Toil(options) as toil:
            with pytest.raises(FailedJobsException):
                toil.start(i)

    def testDeadlockDetection(self) -> None:
        """
        Randomly generate job graphs with various types of cycle in them and
        check they cause an exception properly. Also check that multiple roots
        causes a deadlock exception.
        """
        for test in range(10):
            # Make a random DAG for the set of child edges
            nodeNumber = random.choice(range(2, 20))
            childEdges = self.makeRandomDAG(nodeNumber)
            # Get an adjacency list representation and check is acyclic
            adjacencyList = self.getAdjacencyList(nodeNumber, childEdges)
            assert self.isAcyclic(adjacencyList)

            # Add in follow-on edges - these are returned as a list, and as a set of augmented
            # edges in the adjacency list
            # edges in the adjacency list
            followOnEdges = self.addRandomFollowOnEdges(adjacencyList)
            assert self.isAcyclic(adjacencyList)
            # Make the job graph
            rootJob = self.makeJobGraph(nodeNumber, childEdges, followOnEdges, None)
            rootJob.checkJobGraphAcylic()  # This should not throw an exception
            rootJob.checkJobGraphConnected()  # Nor this
            # Check root detection explicitly
            assert rootJob.getRootJobs() == {rootJob}

            # Test making multiple roots
            childEdges2 = childEdges.copy()
            childEdges2.add(
                (nodeNumber, 1)
            )  # This creates an extra root at "nodeNumber"
            rootJob2 = self.makeJobGraph(
                nodeNumber + 1, childEdges2, followOnEdges, None, False
            )
            with pytest.raises(JobGraphDeadlockException):
                rootJob2.checkJobGraphConnected()

            def checkChildEdgeCycleDetection(fNode: int, tNode: int) -> None:
                childEdges.add((fNode, tNode))  # Create a cycle
                adjacencyList[fNode].add(tNode)
                assert not self.isAcyclic(adjacencyList)
                with pytest.raises(JobGraphDeadlockException):
                    self.makeJobGraph(
                        nodeNumber, childEdges, followOnEdges, None
                    ).checkJobGraphAcylic()
                # Remove the edges
                childEdges.remove((fNode, tNode))
                adjacencyList[fNode].remove(tNode)
                # Check is now acyclic again
                self.makeJobGraph(
                    nodeNumber, childEdges, followOnEdges, None, False
                ).checkJobGraphAcylic()

            def checkFollowOnEdgeCycleDetection(fNode: int, tNode: int) -> None:
                followOnEdges.add((fNode, tNode))  # Create a cycle
                with pytest.raises(JobGraphDeadlockException):
                    self.makeJobGraph(
                        nodeNumber, childEdges, followOnEdges, None, False
                    ).checkJobGraphAcylic()
                # Remove the edges
                followOnEdges.remove((fNode, tNode))
                # Check is now acyclic again
                self.makeJobGraph(
                    nodeNumber, childEdges, followOnEdges, None, False
                ).checkJobGraphAcylic()

            # Now try adding edges that create a cycle

            # Pick a random existing order relationship
            fNode, tNode = self.getRandomEdge(nodeNumber)
            while tNode not in self.reachable(fNode, adjacencyList):
                fNode, tNode = self.getRandomEdge(nodeNumber)

            # Try creating a cycle of child edges
            checkChildEdgeCycleDetection(tNode, fNode)

            # Try adding a self child edge
            node = random.choice(range(nodeNumber))
            checkChildEdgeCycleDetection(node, node)

            # Try adding a follow on edge from a descendant to an ancestor
            checkFollowOnEdgeCycleDetection(tNode, fNode)

            # Try adding a self follow on edge
            checkFollowOnEdgeCycleDetection(node, node)

            # Try adding a follow on edge between two nodes with shared descendants
            fNode, tNode = self.getRandomEdge(nodeNumber)
            if (
                len(
                    self.reachable(tNode, adjacencyList).intersection(
                        self.reachable(fNode, adjacencyList)
                    )
                )
                > 0
                and (fNode, tNode) not in childEdges
                and (fNode, tNode) not in followOnEdges
            ):
                checkFollowOnEdgeCycleDetection(fNode, tNode)

    @slow
    @pytest.mark.slow
    def testNewCheckpointIsLeafVertexNonRootCase(
        self, tmp_path: Path, subtests: SubTests
    ) -> None:
        """
        Test for issue #1465: Detection of checkpoint jobs that are not leaf vertices
        identifies leaf vertices incorrectly

        Test verification of new checkpoint jobs being leaf vertices,
        starting with the following baseline workflow::

            Parent
              |
            Child # Checkpoint=True

        """

        def createWorkflow() -> tuple[Job, FunctionWrappingJob]:
            rootJob = Job.wrapJobFn(simpleJobFn, "Parent")
            childCheckpointJob = rootJob.addChildJobFn(
                simpleJobFn, "Child", checkpoint=True
            )
            return rootJob, childCheckpointJob

        self.runNewCheckpointIsLeafVertexTest(tmp_path, subtests, createWorkflow)

    @slow
    @pytest.mark.slow
    def testNewCheckpointIsLeafVertexRootCase(
        self, tmp_path: Path, subtests: SubTests
    ) -> None:
        """
        Test for issue #1466: Detection of checkpoint jobs that are not leaf vertices
                              omits the workflow root job

        Test verification of a new checkpoint job being leaf vertex,
        starting with a baseline workflow of a single, root job::

            Root # Checkpoint=True

        """

        def createWorkflow() -> tuple[Job, Job]:
            rootJob = Job.wrapJobFn(simpleJobFn, "Root", checkpoint=True)
            return rootJob, rootJob

        self.runNewCheckpointIsLeafVertexTest(tmp_path, subtests, createWorkflow)

    def runNewCheckpointIsLeafVertexTest(
        self,
        tmp_path: Path,
        subtests: SubTests,
        createWorkflowFn: Callable[[], tuple[Job, Job]],
    ) -> None:
        """
        Test verification that a checkpoint job is a leaf vertex using both
        valid and invalid cases.

        :param createWorkflowFn: function to create and new workflow and return a tuple of:

                                 0) the workflow root job
                                 1) a checkpoint job to test within the workflow

        """

        with subtests.test(msg="Test checkpoint job that is a leaf vertex"):
            sub_tmp_path1 = tmp_path / "1"
            sub_tmp_path1.mkdir()
            self.runCheckpointVertexTest(
                *createWorkflowFn(), tmp_path=sub_tmp_path1, expectedException=None
            )

        with subtests.test(
            msg="Test checkpoint job that is not a leaf vertex due to the presence of a service"
        ):
            sub_tmp_path2 = tmp_path / "2"
            sub_tmp_path2.mkdir()
            self.runCheckpointVertexTest(
                *createWorkflowFn(),
                tmp_path=sub_tmp_path2,
                checkpointJobService=TrivialService("LeafTestService"),
                expectedException=JobGraphDeadlockException,
            )

        with subtests.test(
            msg="Test checkpoint job that is not a leaf vertex due to the presence of a child job"
        ):
            sub_tmp_path3 = tmp_path / "3"
            sub_tmp_path3.mkdir()
            self.runCheckpointVertexTest(
                *createWorkflowFn(),
                tmp_path=sub_tmp_path3,
                checkpointJobChild=Job.wrapJobFn(simpleJobFn, "LeafTestChild"),
                expectedException=JobGraphDeadlockException,
            )

        with subtests.test(
            msg="Test checkpoint job that is not a leaf vertex due to the presence of a follow-on job"
        ):
            sub_tmp_path4 = tmp_path / "4"
            sub_tmp_path4.mkdir()
            self.runCheckpointVertexTest(
                *createWorkflowFn(),
                tmp_path=sub_tmp_path4,
                checkpointJobFollowOn=Job.wrapJobFn(simpleJobFn, "LeafTestFollowOn"),
                expectedException=JobGraphDeadlockException,
            )

    def runCheckpointVertexTest(
        self,
        workflowRootJob: Job,
        checkpointJob: Job,
        tmp_path: Path,
        checkpointJobService: Optional[Job.Service] = None,
        checkpointJobChild: Optional[Job] = None,
        checkpointJobFollowOn: Optional[Job] = None,
        expectedException: Optional[type[Exception]] = None,
    ) -> None:
        """
        Modifies the checkpoint job according to the given parameters
        then runs the workflow, checking for the expected exception, if any.
        """

        assert checkpointJob.checkpoint

        if checkpointJobService is not None:
            checkpointJob.addService(checkpointJobService)
        if checkpointJobChild is not None:
            checkpointJob.addChild(checkpointJobChild)
        if checkpointJobFollowOn is not None:
            checkpointJob.addFollowOn(checkpointJobFollowOn)

        # Run the workflow and check for the expected behavior
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.logLevel = "INFO"
        if expectedException is None:
            Job.Runner.startToil(workflowRootJob, options)
        else:
            with pytest.raises(expectedException):
                Job.Runner.startToil(workflowRootJob, options)

    @slow
    @pytest.mark.slow
    def testEvaluatingRandomDAG(self, tmp_path: Path) -> None:
        """
        Randomly generate test input then check that the job graph can be
        run successfully, using the existence of promises
        to validate the run.
        """
        jobStore = tmp_path / "jobstore"
        for test in range(5):
            # Temporary file
            tempDir = tmp_path / f"tempDir{test}"
            tempDir.mkdir()
            # Make a random DAG for the set of child edges
            nodeNumber = random.choice(range(2, 8))
            childEdges = self.makeRandomDAG(nodeNumber)
            # Get an adjacency list representation and check is acyclic
            adjacencyList = self.getAdjacencyList(nodeNumber, childEdges)
            assert self.isAcyclic(adjacencyList)
            # Add in follow on edges - these are returned as a list, and as a set of augmented
            # edges in the adjacency list
            followOnEdges = self.addRandomFollowOnEdges(adjacencyList)
            assert self.isAcyclic(adjacencyList)
            # Make the job graph
            rootJob = self.makeJobGraph(nodeNumber, childEdges, followOnEdges, tempDir)
            # Run the job  graph
            options = Job.Runner.getDefaultOptions("%s.%i" % (jobStore, test))
            options.logLevel = "DEBUG"
            options.retryCount = 1
            options.badWorker = 0.25
            options.badWorkerFailInterval = 0.01
            # Because we're going to be killing the services all the time for
            # restarts, make sure they are paying attention.
            options.servicePollingInterval = 1

            # Now actually run the workflow
            try:
                with Toil(options) as toil:
                    toil.start(rootJob)
                numberOfFailedJobs = 0
            except FailedJobsException as e:
                numberOfFailedJobs = e.numberOfFailedJobs

            # Restart until successful or failed
            totalTrys = 1
            options.restart = True
            while numberOfFailedJobs != 0:
                try:
                    with Toil(options) as toil:
                        toil.restart()
                    numberOfFailedJobs = 0
                except FailedJobsException as e:
                    numberOfFailedJobs = e.numberOfFailedJobs
                    if totalTrys > 32:  # p(fail after this many restarts) ~= 0.5**32
                        pytest.fail("Exceeded a reasonable number of restarts")
                    totalTrys += 1

            # For each job check it created a valid output file and add the ordering
            # relationships contained within the output file to the ordering relationship,
            # so we can check they are compatible with the relationships defined by the job DAG.
            ordering = None
            for i in range(nodeNumber):
                with (tempDir / str(i)).open() as fH:
                    ordering = list(map(int, fH.readline().split()))
                    assert int(ordering[-1]) == i
                    for j in ordering[:-1]:
                        adjacencyList[int(j)].add(i)
            # Check the ordering retains an acyclic graph
            if not self.isAcyclic(adjacencyList):
                print("ORDERING", ordering)
                print("CHILD EDGES", childEdges)
                print("FOLLOW ON EDGES", followOnEdges)
                print("ADJACENCY LIST", adjacencyList)
            assert self.isAcyclic(adjacencyList)

    @staticmethod
    def getRandomEdge(nodeNumber: int) -> tuple[int, int]:
        assert nodeNumber > 1
        fNode = random.choice(range(nodeNumber - 1))
        return fNode, random.choice(range(fNode + 1, nodeNumber))

    @staticmethod
    def makeRandomDAG(nodeNumber: int) -> set[tuple[int, int]]:
        """
        Makes a random dag with "nodeNumber" nodes in which all nodes are connected. Return value
        is list of edges, each of form (a, b), where a and b are integers >= 0 < nodeNumber
        referring to nodes and the edge is from a to b.
        """
        # Pick number of total edges to create
        edgeNumber = random.choice(
            range(nodeNumber - 1, 1 + (nodeNumber * (nodeNumber - 1) // 2))
        )
        # Make a spanning tree of edges so that nodes are connected
        edges = {(random.choice(range(i)), i) for i in range(1, nodeNumber)}
        # Add extra random edges until there are edgeNumber edges
        while len(edges) < edgeNumber:
            edges.add(TestJob.getRandomEdge(nodeNumber))
        return edges

    @staticmethod
    def getAdjacencyList(
        nodeNumber: int, edges: set[tuple[int, int]]
    ) -> list[set[int]]:
        """
        Make adjacency list representation of edges
        """
        adjacencyList: list[set[int]] = [set() for _ in range(nodeNumber)]
        for fNode, tNode in edges:
            adjacencyList[fNode].add(tNode)
        return adjacencyList

    def reachable(self, node: int, adjacencyList: list[set[int]]) -> set[int]:
        """
        Find the set of nodes reachable from this node (including the node). Return is a set of
        integers.
        """
        visited: set[int] = set()

        def dfs(fNode: int) -> None:
            if fNode not in visited:
                visited.add(fNode)
                list(map(dfs, adjacencyList[fNode]))

        dfs(node)
        return visited

    def addRandomFollowOnEdges(
        self, childAdjacencyList: list[set[int]]
    ) -> set[tuple[int, int]]:
        """
        Adds random follow on edges to the graph, represented as an adjacency list. The follow on
        edges are returned as a set and their augmented edges are added to the adjacency list.
        """

        def makeAugmentedAdjacencyList() -> list[set[int]]:
            augmentedAdjacencyList = [
                childAdjacencyList[i].union(followOnAdjacencyList[i])
                for i in range(len(childAdjacencyList))
            ]

            def addImpliedEdges(node: int, followOnEdges: set[int]) -> None:
                # Let node2 be a child of node or a successor of a child of node.
                # For all node2 the following adds an edge to the augmented
                # adjacency list from node2 to each followOn of node

                visited: set[int] = set()

                def f(node2: int) -> None:
                    if node2 not in visited:
                        visited.add(node2)
                        for i in followOnEdges:
                            augmentedAdjacencyList[node2].add(i)
                        list(map(f, childAdjacencyList[node2]))
                        list(map(f, followOnAdjacencyList[node2]))

                list(map(f, childAdjacencyList[node]))

            for node in range(len(followOnAdjacencyList)):
                addImpliedEdges(node, followOnAdjacencyList[node])
            return augmentedAdjacencyList

        followOnEdges = set()
        followOnAdjacencyList: list[set[int]] = [set() for i in childAdjacencyList]
        # Loop to create the follow on edges (try 1000 times)
        while random.random() > 0.001:
            fNode, tNode = TestJob.getRandomEdge(len(childAdjacencyList))

            # Make an adjacency list including augmented edges and proposed
            # follow on edge

            # Add the new follow on edge
            followOnAdjacencyList[fNode].add(tNode)

            augmentedAdjacencyList = makeAugmentedAdjacencyList()

            # If the augmented adjacency doesn't contain a cycle then add the follow on edge to
            # the list of follow ons else remove the follow on edge from the follow on adjacency
            # list.
            if self.isAcyclic(augmentedAdjacencyList):
                followOnEdges.add((fNode, tNode))
            else:
                followOnAdjacencyList[fNode].remove(tNode)

        # Update adjacency list adding in augmented edges
        childAdjacencyList[:] = makeAugmentedAdjacencyList()[:]

        return followOnEdges

    def makeJobGraph(
        self,
        nodeNumber: int,
        childEdges: set[tuple[int, int]],
        followOnEdges: set[tuple[int, int]],
        outPath: Optional[Path],
        addServices: bool = True,
    ) -> Job:
        """
        Converts a DAG into a job graph. childEdges and followOnEdges are the lists of child and
        followOn edges.
        """
        # Map of jobs to the list of promises they have
        jobsToPromisesMap: dict[FunctionWrappingJob, list[Promise]] = {}

        def makeJob(string: str) -> FunctionWrappingJob:
            promises: list[Promise] = []
            job = Job.wrapFn(
                fn2Test,
                promises,
                string,
                None if outPath is None else os.path.join(outPath, string),
                cores=0.1,
                memory="0.5G",
                disk="0.1G",
            )
            jobsToPromisesMap[job] = promises
            return job

        # Make the jobs
        jobs = [makeJob(str(i)) for i in range(nodeNumber)]

        # Record predecessors for sampling
        predecessors: dict[FunctionWrappingJob, list[FunctionWrappingJob]] = (
            collections.defaultdict(list)
        )

        # Make the edges
        for fNode, tNode in childEdges:
            jobs[fNode].addChild(jobs[tNode])
            predecessors[jobs[tNode]].append(jobs[fNode])
        for fNode, tNode in followOnEdges:
            jobs[fNode].addFollowOn(jobs[tNode])
            predecessors[jobs[tNode]].append(jobs[fNode])

        # Map of jobs to return values
        jobsToRvs = {
            job: (
                job.addService(
                    TrivialService(
                        cast(str, job.rv()), cores=0.1, memory="0.5G", disk="0.1G"
                    )
                )
                if addServices
                else job.rv()
            )
            for job in jobs
        }

        def getRandomPredecessor(job: FunctionWrappingJob) -> FunctionWrappingJob:
            predecessor = random.choice(list(predecessors[job]))
            while random.random() > 0.5 and len(predecessors[predecessor]) > 0:
                predecessor = random.choice(list(predecessors[predecessor]))
            return predecessor

        # Connect up set of random promises compatible with graph
        while random.random() > 0.01:
            job = random.choice(list(jobsToPromisesMap.keys()))
            promises = jobsToPromisesMap[job]
            if len(predecessors[job]) > 0:
                predecessor = getRandomPredecessor(job)
                promises.append(jobsToRvs[predecessor])

        return jobs[0]

    def isAcyclic(self, adjacencyList: list[set[int]]) -> bool:
        """
        Returns true if there are any cycles in the graph, which is represented as an adjacency
        list.
        """

        def cyclic(fNode: int, visited: set[int], stack: list[int]) -> Union[bool, int]:
            if fNode not in visited:
                visited.add(fNode)
                assert fNode not in stack
                stack.append(fNode)
                for tNode in adjacencyList[fNode]:
                    if cyclic(tNode, visited, stack):
                        return True
                assert stack.pop() == fNode
            return fNode in stack

        visited: set[int] = set()
        for i in range(len(adjacencyList)):
            if cyclic(i, visited, []):
                return False
        return True


def simpleJobFn(job: ServiceHostJob, value: str) -> None:
    job.fileStore.log_to_leader(value)


def fn1Test(string: str, outputFile: FileDescriptorOrPath) -> str:
    """
    Function appends the next character after the last character in the given
    string to the string, writes the string to a file, and returns it. For
    example, if string is "AB", we will write and return "ABC".
    """

    rV = string + chr(ord(string[-1]) + 1)
    with open(outputFile, "w") as fH:
        fH.write(rV)
    return rV


def fn2Test(pStrings: list[str], s: str, outputFile: Path) -> str:
    """
    Function concatenates the strings in pStrings and s, in that order, and writes the result to
    the output file. Returns s.
    """
    with open(outputFile, "w") as fH:
        fH.write(" ".join(pStrings) + " " + s)
    return s


def trivialParent(job: Job) -> None:
    strandedJob = JobFunctionWrappingJob(child)
    failingJob = JobFunctionWrappingJob(errorChild)

    job.addChild(failingJob)
    job.addChild(strandedJob)
    failingJob.addChild(strandedJob)


def parent(job: Job) -> None:
    childJob = JobFunctionWrappingJob(child)
    strandedJob = JobFunctionWrappingJob(child)
    failingJob = JobFunctionWrappingJob(errorChild)

    job.addChild(childJob)
    job.addChild(strandedJob)
    childJob.addChild(failingJob)
    failingJob.addChild(strandedJob)


def diamond(job: Job) -> None:
    childJob = JobFunctionWrappingJob(child)
    strandedJob = JobFunctionWrappingJob(child)
    failingJob = JobFunctionWrappingJob(errorChild)

    job.addChild(childJob)
    job.addChild(failingJob)
    childJob.addChild(strandedJob)
    failingJob.addChild(strandedJob)


def child(job: Job) -> None:
    assert job.cores is not None
    assert job.disk is not None
    assert job.memory is not None
    assert job.preemptible is not None


def errorChild(job: Job) -> NoReturn:
    raise RuntimeError("Child failure")


class TrivialService(Job.Service):
    def __init__(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Service that does nothing, used to check for deadlocks"""
        Job.Service.__init__(self, *args, **kwargs)
        self.message = message

    def start(self, job: ServiceHostJob) -> str:
        return self.message

    def stop(self, job: ServiceHostJob) -> None:
        pass

    def check(self) -> bool:
        return True



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Optional

from toil.batchSystems.abstractBatchSystem import (
    BatchSystemSupport,
    UpdatedBatchJobInfo,
)
from toil.batchSystems.singleMachine import SingleMachineBatchSystem
from toil.common import Config
from toil.job import JobDescription
from toil.lib.threading import cpu_count

logger = logging.getLogger(__name__)


class BatchSystemLocalSupport(BatchSystemSupport):
    """Adds a local queue for helper jobs, useful for CWL & others."""

    def __init__(
        self, config: Config, maxCores: float, maxMemory: float, maxDisk: int
    ) -> None:
        super().__init__(config, maxCores, maxMemory, maxDisk)
        max_local_jobs = (
            config.max_local_jobs if config.max_local_jobs is not None else cpu_count()
        )
        self.localBatch: SingleMachineBatchSystem = SingleMachineBatchSystem(
            config, maxCores, maxMemory, maxDisk, max_jobs=max_local_jobs
        )

    def handleLocalJob(self, command: str, jobDesc: JobDescription) -> Optional[int]:
        """
        To be called by issueBatchJob.

        Returns the jobID if the jobDesc has been submitted to the local queue,
        otherwise returns None
        """
        if not self.config.run_local_jobs_on_workers and jobDesc.local:
            # Since singleMachine.py doesn't typecheck yet and MyPy is ignoring
            # it, it will raise errors here unless we add type annotations to
            # everything we get back from it. The easiest way to do that seems
            # to be to put it in a variable with a type annotation on it. That
            # somehow doesn't error whereas just returning the value complains
            # we're returning an Any. TODO: When singleMachine.py typechecks,
            # remove all these extra variables.
            local_id: int = self.localBatch.issueBatchJob(command, jobDesc)
            return local_id
        else:
            return None

    def killLocalJobs(self, jobIDs: list[int]) -> None:
        """
        Will kill all local jobs that match the provided jobIDs.

        To be called by killBatchJobs.
        """
        self.localBatch.killBatchJobs(jobIDs)

    def getIssuedLocalJobIDs(self) -> list[int]:
        """To be called by getIssuedBatchJobIDs."""
        local_ids: list[int] = self.localBatch.getIssuedBatchJobIDs()
        return local_ids

    def getRunningLocalJobIDs(self) -> dict[int, float]:
        """To be called by getRunningBatchJobIDs()."""
        local_running: dict[int, float] = self.localBatch.getRunningBatchJobIDs()
        return local_running

    def getUpdatedLocalJob(self, maxWait: int) -> Optional[UpdatedBatchJobInfo]:
        """To be called by getUpdatedBatchJob()."""
        return self.localBatch.getUpdatedBatchJob(maxWait)

    def getNextJobID(self) -> int:
        """Must be used to get job IDs so that the local and batch jobs do not conflict."""
        # TODO: This reaches deep into SingleMachineBatchSystem when it probably shouldn't
        with self.localBatch.jobIndexLock:
            jobID: int = self.localBatch.jobIndex
            self.localBatch.jobIndex += 1
        return jobID

    def shutdownLocal(self) -> None:
        """To be called from shutdown()."""
        self.localBatch.shutdown()



import json
import logging
import os
import re
import shutil
import string
import subprocess
import unittest
from collections.abc import Generator
from pathlib import Path
from typing import Any, Optional, Union, cast
from unittest.mock import patch
from uuid import uuid4

import pytest
import WDL.Error
import WDL.Expr

from toil.fileStores import FileID
from toil.test import (
    ToilTest,
    get_data,
    needs_docker,
    needs_docker_cuda,
    needs_google_storage,
    needs_singularity_or_docker,
    needs_wdl,
    slow,
)
from toil.version import exactPython
from toil.wdl.wdltoil import (
    WDLSectionJob,
    WDLWorkflowGraph,
    parse_disks,
    remove_common_leading_whitespace,
)

logger = logging.getLogger(__name__)


WDL_CONFORMANCE_TEST_REPO = "https://github.com/DataBiosphere/wdl-conformance-tests.git"
WDL_CONFORMANCE_TEST_COMMIT = "baf44bcc7e6f6927540adf77d91b26a5558ae4b7"
# These tests are known to require things not implemented by
# Toil and will not be run in CI.
WDL_CONFORMANCE_TESTS_UNSUPPORTED_BY_TOIL = [
    16,  # Basic object test (deprecated and removed in 1.1); MiniWDL and toil-wdl-runner do not support Objects, so this will fail if ran by them
    21,  # Parser: expression placeholders in strings in conditional expressions in 1.0, Cromwell style; Fails with MiniWDL and toil-wdl-runner
    64,  # Legacy test for as_map_as_input; It looks like MiniWDL does not have the function as_map()
    77,  # Test that array cannot coerce to a string. WDL 1.1 does not allow compound types to coerce into a string. This should return a TypeError.
]
WDL_UNIT_TESTS_UNSUPPORTED_BY_TOIL = [
    14,  # test_object, Objects are not supported
    19,  # map_to_struct, miniwdl cannot coerce map to struct, https://github.com/chanzuckerberg/miniwdl/issues/712
    52,  # relative_and_absolute, needs root to run
    58,  # test_gpu, needs gpu to run, else warning
    59,  # will be fixed in #5001
    66,  # This needs way too many resources (and actually doesn't work?), see https://github.com/DataBiosphere/wdl-conformance-tests/blob/2d617b703a33791f75f30a9db43c3740a499cd89/README_UNIT.md?plain=1#L8
    67,  # same as above
    68,  # Bug, see #https://github.com/DataBiosphere/toil/issues/4993
    69,  # Same as 68
    87,  # MiniWDL does not handle metacharacters properly when running regex, https://github.com/chanzuckerberg/miniwdl/issues/709
    97,  # miniwdl bug, see https://github.com/chanzuckerberg/miniwdl/issues/701
    105,  # miniwdl (and toil) bug, unserializable json is serialized, see https://github.com/chanzuckerberg/miniwdl/issues/702
    107,  # object not supported
    108,  # object not supported
    109,  # object not supported
    110,  # object not supported
    120,  # miniwdl bug, see https://github.com/chanzuckerberg/miniwdl/issues/699
    131,  # miniwdl bug, evalerror, see https://github.com/chanzuckerberg/miniwdl/issues/700
    134,  # same as 131
    144,  # miniwdl and toil bug
]


@pytest.fixture(scope="function")
def wdl_conformance_test_repo(tmp_path: Path) -> Generator[Path]:
    try:
        p = subprocess.Popen(
            f"git clone {WDL_CONFORMANCE_TEST_REPO} {str(tmp_path)} && cd {str(tmp_path)} && git checkout {WDL_CONFORMANCE_TEST_COMMIT}",
            shell=True,
        )

        p.communicate()

        if p.returncode > 0:
            raise RuntimeError("Could not clone WDL conformance tests")
        yield tmp_path
    finally:
        pass  # no cleanup needed


class TestWDLConformance:
    """
    WDL conformance tests for Toil.
    """

    def check(self, p: "subprocess.CompletedProcess[bytes]") -> None:
        """
        Make sure a call completed or explain why it failed.
        """

        if p.returncode != 0:
            logger.error(
                "Failed process standard output: %s",
                p.stdout.decode("utf-8", errors="replace"),
            )
            logger.error(
                "Failed process standard error: %s",
                p.stderr.decode("utf-8", errors="replace"),
            )

        p.check_returncode()

    @slow
    def test_unit_tests_v11(self, wdl_conformance_test_repo: Path) -> None:
        # There are still some bugs with the WDL spec, use a fixed version until
        # See comments of https://github.com/openwdl/wdl/pull/669
        os.chdir(wdl_conformance_test_repo)
        repo_url = "https://github.com/stxue1/wdl.git"
        repo_branch = "wdl-1.1.3-fixes"
        commands1 = [
            exactPython,
            "setup_unit_tests.py",
            "-v",
            "1.1",
            "--extra-patch-data",
            "unit_tests_patch_data.yaml",
            "--repo",
            repo_url,
            "--branch",
            repo_branch,
            "--force-pull",
        ]
        p1 = subprocess.run(commands1, capture_output=True)
        self.check(p1)
        commands2 = [
            exactPython,
            "run_unit.py",
            "-r",
            "toil-wdl-runner",
            "-v",
            "1.1",
            "--progress",
            "--exclude-numbers",
            ",".join([str(t) for t in WDL_UNIT_TESTS_UNSUPPORTED_BY_TOIL]),
        ]
        p2 = subprocess.run(commands2, capture_output=True)
        self.check(p2)

    # estimated running time: 10 minutes
    @slow
    def test_conformance_tests_v10(self, wdl_conformance_test_repo: Path) -> None:
        os.chdir(wdl_conformance_test_repo)
        commands = [
            exactPython,
            "run.py",
            "--runner",
            "toil-wdl-runner",
            "--conformance-file",
            "conformance.yaml",
            "-v",
            "1.0",
        ]
        if WDL_CONFORMANCE_TESTS_UNSUPPORTED_BY_TOIL:
            commands.append("--exclude-numbers")
            commands.append(
                ",".join([str(t) for t in WDL_CONFORMANCE_TESTS_UNSUPPORTED_BY_TOIL])
            )
        p = subprocess.run(commands, capture_output=True)

        self.check(p)

    # estimated running time: 10 minutes
    @slow
    def test_conformance_tests_v11(self, wdl_conformance_test_repo: Path) -> None:
        os.chdir(wdl_conformance_test_repo)
        commands = [
            exactPython,
            "run.py",
            "--runner",
            "toil-wdl-runner",
            "--conformance-file",
            "conformance.yaml",
            "-v",
            "1.1",
        ]
        if WDL_CONFORMANCE_TESTS_UNSUPPORTED_BY_TOIL:
            commands.append("--exclude-numbers")
            commands.append(
                ",".join([str(t) for t in WDL_CONFORMANCE_TESTS_UNSUPPORTED_BY_TOIL])
            )
        p = subprocess.run(commands, capture_output=True)

        self.check(p)

    @slow
    def test_conformance_tests_integration(
        self, wdl_conformance_test_repo: Path
    ) -> None:
        os.chdir(wdl_conformance_test_repo)
        commands = [
            exactPython,
            "run.py",
            "--runner",
            "toil-wdl-runner",
            "-v",
            "1.0",
            "--conformance-file",
            "integration.yaml",
            "--id",
            "encode,tut01,tut02,tut03,tut04",
        ]
        p = subprocess.run(
            commands,
            capture_output=True,
        )

        self.check(p)


class TestWDL:
    """Tests for Toil's MiniWDL-based implementation."""

    base_command = [exactPython, "-m", "toil.wdl.wdltoil"]

    # We inherit a testMD5sum but it is going to need Singularity or Docker
    # now. And also needs to have a WDL 1.0+ WDL file. So we replace it.
    @needs_singularity_or_docker
    def test_MD5sum(self, tmp_path: Path) -> None:
        """Test if Toil produces the same outputs as known good outputs for WDL's
        GATK tutorial #1."""
        with get_data("test/wdl/md5sum/md5sum.1.0.wdl") as wdl:
            with get_data("test/wdl/md5sum/md5sum.json") as json_file:
                result_json = subprocess.check_output(
                    self.base_command
                    + [
                        str(wdl),
                        str(json_file),
                        "-o",
                        str(tmp_path),
                        "--logDebug",
                        "--retryCount=0",
                    ]
                )
                result = json.loads(result_json)

                assert "ga4ghMd5.value" in result
                assert isinstance(result["ga4ghMd5.value"], str)
                assert os.path.exists(result["ga4ghMd5.value"])
                assert os.path.basename(result["ga4ghMd5.value"]) == "md5sum.txt"

    def test_url_to_file(self, tmp_path: Path) -> None:
        """
        Test if web URL strings can be coerced to usable Files.
        """
        with get_data("test/wdl/testfiles/url_to_file.wdl") as wdl:
            result_json = subprocess.check_output(
                self.base_command
                + [str(wdl), "-o", str(tmp_path), "--logInfo", "--retryCount=0"]
            )
            result = json.loads(result_json)

            assert "url_to_file.first_line" in result
            assert isinstance(result["url_to_file.first_line"], str)
            assert result["url_to_file.first_line"] == "chr1\t248387328"

    def test_string_file_coercion(self, tmp_path: Path) -> None:
        """
        Test if input Files can be coerced to string and back.
        """
        with get_data("test/wdl/testfiles/string_file_coercion.wdl") as wdl:
            with get_data("test/wdl/testfiles/string_file_coercion.json") as json_file:
                result_json = subprocess.check_output(
                    self.base_command
                    + [
                        str(wdl),
                        str(json_file),
                        "-o",
                        str(tmp_path),
                        "--logInfo",
                        "--retryCount=0"
                    ]
                )
                result = json.loads(result_json)

                assert "StringFileCoercion.output_file" in result

    @needs_docker
    def test_wait(self, tmp_path: Path) -> None:
        """
        Test if Bash "wait" works in WDL scripts.
        """
        with get_data("test/wdl/testfiles/wait.wdl") as wdl:
            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path),
                    "--logInfo",
                    "--retryCount=0",
                    "--wdlContainer=docker",
                ]
            )
            result = json.loads(result_json)

            assert "wait.result" in result
            assert isinstance(result["wait.result"], str)
            assert result["wait.result"] == "waited"

    @needs_singularity_or_docker
    def test_workflow_file_deletion(self, tmp_path: Path) -> None:
        """
        Test if Toil can delete non-output outputs at the end of a workflow.
        """
        # Keep a job store around to inspect for files.
        (tmp_path / "jobStore").mkdir()
        job_store = tmp_path / "jobStore" / "tree"

        # Make a working directory to run in
        work_dir = tmp_path / "workDir"
        work_dir.mkdir()
        # Make the file that will be imported from a string in the workflow
        referenced_file = work_dir / "localfile.txt"
        with referenced_file.open("w") as f:
            f.write("This file is imported by local path in the workflow")
        # Make the file to pass as input
        sent_in_file = work_dir / "sent_in.txt"
        with sent_in_file.open("w") as f:
            f.write("This file is sent in as input")

        with get_data("test/wdl/testfiles/drop_files.wdl") as wdl:
            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path / "output"),
                    "--jobStore",
                    job_store,
                    "--clean=never",
                    "--logInfo",
                    "--retryCount=0",
                    '--input={"file_in": "' + str(sent_in_file) + '"}',
                ],
                cwd=work_dir,
            )
            result = json.loads(result_json)

            # Get all the file values in the job store.
            all_file_values = set()
            for directory, _, files in os.walk(
                job_store
            ):  # can't switch to job_store.walk() until Python 3.12 is the minimum version
                for filename in files:
                    with (Path(directory) / filename).open(
                        encoding="utf-8", errors="replace"
                    ) as f:
                        all_file_values.add(f.read().rstrip())

            # Make sure the files with the right contents are in the job store and
            # the files with the wrong contents aren't anymore.
            #
            # This assumes no top-level cleanup in the main driver script.

            # These are all created inside the workflow and not output
            assert (
                "This file is imported by local path in the workflow"
                not in all_file_values
            )
            assert "This file is consumed by a task call" not in all_file_values
            assert (
                "This file is created in a task inputs section" not in all_file_values
            )
            assert "This file is created in a runtime section" not in all_file_values
            assert "This task file is not used" not in all_file_values
            assert "This file should be discarded" not in all_file_values
            assert "This file is dropped by a subworkflow" not in all_file_values
            assert "This file gets stored in a variable" not in all_file_values
            assert "This file never gets stored in a variable" not in all_file_values

            # These are created inside the workflow and output
            assert "3" in all_file_values
            assert "This file is collected as a task output twice" in all_file_values
            assert "This file should be kept" in all_file_values
            assert "This file is kept by a subworkflow" in all_file_values

            # These are sent into the workflow from the enclosing environment and
            # should not be deleted.
            assert "This file is sent in as input" in all_file_values

            # Make sure we didn't somehow delete the file sent as input
            assert sent_in_file.exists()

    @needs_singularity_or_docker
    def test_all_call_outputs(self, tmp_path: Path) -> None:
        """
        Test if Toil can collect all call outputs from a workflow that doesn't expose them.
        """
        with get_data("test/wdl/testfiles/not_enough_outputs.wdl") as wdl:
            # With no flag we don't include the call outputs
            result_json = subprocess.check_output(
                self.base_command
                + [str(wdl), "-o", str(tmp_path), "--logInfo", "--retryCount=0"]
            )
            result = json.loads(result_json)

            assert "wf.only_result" in result
            assert "wf.do_math.square" not in result
            assert "wf.do_math.cube" not in result
            assert "wf.should_never_output" not in result

            # With flag off we don't include the call outputs
            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path),
                    "--logInfo",
                    "--retryCount=0",
                    "--allCallOutputs=false",
                ]
            )
            result = json.loads(result_json)

            assert "wf.only_result" in result
            assert "wf.do_math.square" not in result
            assert "wf.do_math.cube" not in result
            assert "wf.should_never_output" not in result

            # With flag on we do include the call outputs
            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path),
                    "--logInfo",
                    "--retryCount=0",
                    "--allCallOutputs=on",
                ]
            )
            result = json.loads(result_json)

            assert "wf.only_result" in result
            assert "wf.do_math.square" in result
            assert "wf.do_math.cube" in result
            assert "wf.should_never_output" not in result

    @needs_singularity_or_docker
    def test_croo_detection(self, tmp_path: Path) -> None:
        """
        Test if Toil can detect and do something sensible with Cromwell Output Organizer workflows.
        """
        with get_data("test/wdl/testfiles/croo.wdl") as wdl:
            # With no flag we should include all task outputs
            result_json = subprocess.check_output(
                self.base_command
                + [str(wdl), "-o", str(tmp_path), "--logInfo", "--retryCount=0"]
            )
            result = json.loads(result_json)

            assert "wf.only_result" in result
            assert "wf.do_math.square" in result
            assert "wf.do_math.cube" in result
            assert "wf.should_never_output" not in result

            # With flag off we obey the WDL spec even if we're suspicious
            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path),
                    "--logInfo",
                    "--retryCount=0",
                    "--allCallOutputs=off",
                ]
            )
            result = json.loads(result_json)

            assert "wf.only_result" in result
            assert "wf.do_math.square" not in result
            assert "wf.do_math.cube" not in result
            assert "wf.should_never_output" not in result

    @needs_singularity_or_docker
    def test_caching(self, tmp_path: Path) -> None:
        """
        Test if Toil can cache task runs.
        """
        with get_data("test/wdl/testfiles/random.wdl") as wdl:
            cachedir = tmp_path / "cache"
            cachedir.mkdir()
            caching_env = dict(os.environ)
            caching_env["MINIWDL__CALL_CACHE__GET"] = "true"
            caching_env["MINIWDL__CALL_CACHE__PUT"] = "true"
            caching_env["MINIWDL__CALL_CACHE__DIR"] = str(cachedir)

            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path / "out1"),
                    "--logInfo",
                    "--retryCount=0",
                    '--inputs={"random.task_1_input": 1, "random.task_2_input": 1}',
                ],
                env=caching_env,
            )
            result_initial = json.loads(result_json)

            assert "random.value_seen" in result_initial
            assert "random.value_written" in result_initial

            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path / "out2"),
                    "--logInfo",
                    "--retryCount=0",
                    '--inputs={"random.task_1_input": 1, "random.task_2_input": 1}',
                ],
                env=caching_env,
            )
            result_cached = json.loads(result_json)

            assert "random.value_seen" in result_cached
            assert "random.value_written" in result_cached

            assert (
                result_cached["random.value_seen"]
                == result_initial["random.value_seen"]
            )
            assert (
                result_cached["random.value_written"]
                == result_initial["random.value_written"]
            )

            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path / "out3"),
                    "--logInfo",
                    "--retryCount=0",
                    '--inputs={"random.task_1_input": 2, "random.task_2_input": 1}',
                ],
                env=caching_env,
            )
            result_not_cached = json.loads(result_json)

            assert "random.value_seen" in result_not_cached
            assert "random.value_written" in result_not_cached

            assert (
                result_not_cached["random.value_seen"]
                != result_initial["random.value_seen"]
            )
            assert (
                result_not_cached["random.value_written"]
                != result_initial["random.value_written"]
            )

            result_json = subprocess.check_output(
                self.base_command
                + [
                    str(wdl),
                    "-o",
                    str(tmp_path / "out4"),
                    "--logInfo",
                    "--retryCount=0",
                    '--inputs={"random.task_1_input": 1, "random.task_2_input": 2}',
                ],
                env=caching_env,
            )
            result_part_cached = json.loads(result_json)

            assert "random.value_seen" in result_part_cached
            assert "random.value_written" in result_part_cached

            assert (
                result_part_cached["random.value_seen"]
                == result_initial["random.value_seen"]
            )
            assert (
                result_part_cached["random.value_written"]
                != result_initial["random.value_written"]
            )
            assert (
                result_part_cached["random.value_written"]
                != result_not_cached["random.value_written"]
            )

    def test_url_to_optional_file(self, tmp_path: Path) -> None:
        """
        Test if missing and error-producing URLs are handled correctly for optional File? values.
        """
        with get_data("test/wdl/testfiles/url_to_optional_file.wdl") as wdl:

            def run_for_code(code: int) -> dict[str, Any]:
                """
                Run a workflow coercing URL to File? where the URL returns the given status code.

                Return the parsed output.
                """
                logger.info("Test optional file with HTTP code %s", code)
                json_value = '{"url_to_optional_file.http_code": %d}' % code
                result_json = subprocess.check_output(
                    self.base_command
                    + [
                        str(wdl),
                        json_value,
                        "-o",
                        str(tmp_path),
                        "--logInfo",
                        "--retryCount=0",
                    ]
                )
                result = json.loads(result_json)
                return cast(dict[str, Any], json.loads(result_json))

            # Check files that exist
            result = run_for_code(200)
            assert "url_to_optional_file.out_file" in result
            assert result["url_to_optional_file.out_file"] is not None

            for code in (404, 410):
                # Check files that definitely don't
                result = run_for_code(code)
                assert "url_to_optional_file.out_file" in result
                assert result["url_to_optional_file.out_file"] is None

            for code in (402, 418, 500, 502):
                # Check that cases where the server refuses to say if the file
                # exists stop the workflow.
                with pytest.raises(subprocess.CalledProcessError):
                    run_for_code(code)

    def test_missing_output_directory(self, tmp_path: Path) -> None:
        """
        Test if Toil can run a WDL workflow into a new directory.
        """
        with get_data("test/wdl/md5sum/md5sum.1.0.wdl") as wdl:
            with get_data("test/wdl/md5sum/md5sum.json") as json_file:
                subprocess.check_call(
                    self.base_command
                    + [
                        str(wdl),
                        str(json_file),
                        "-o",
                        str(tmp_path / "does" / "not" / "exist"),
                        "--logDebug",
                        "--retryCount=0",
                    ]
                )

    @needs_singularity_or_docker
    def test_miniwdl_self_test(
        self, tmp_path: Path, extra_args: Optional[list[str]] = None
    ) -> None:
        """Test if the MiniWDL self test runs and produces the expected output."""
        with get_data("test/wdl/miniwdl_self_test/self_test.wdl") as wdl_file:
            with get_data("test/wdl/miniwdl_self_test/inputs.json") as json_file:

                result_json = subprocess.check_output(
                    self.base_command
                    + [
                        str(wdl_file),
                        str(json_file),
                        "--logDebug",
                        "-o",
                        str(tmp_path),
                        "--outputDialect",
                        "miniwdl",
                    ]
                    + (extra_args or [])
                )
                result = json.loads(result_json)

                # Expect MiniWDL-style output with a designated "dir"

                assert "dir" in result
                assert isinstance(result["dir"], str)
                out_dir = result["dir"]

                assert "outputs" in result
                assert isinstance(result["outputs"], dict)
                outputs = result["outputs"]

                assert "hello_caller.message_files" in outputs
                assert isinstance(outputs["hello_caller.message_files"], list)
                assert len(outputs["hello_caller.message_files"]) == 2
                for item in outputs["hello_caller.message_files"]:
                    # All the files should be strings in the "out" directory
                    assert isinstance(item, str), "File output must be a string"
                    assert item.startswith(
                        out_dir
                    ), "File output must be in the output directory"

                    # Look at the filename within that directory
                    name_in_out_dir = item[len(out_dir) :]

                    # Ity should contain the job name of "hello", so they are human-readable.
                    assert (
                        "hello" in name_in_out_dir
                    ), f"File output {name_in_out_dir} should have the originating task name in it"

                    # And it should not contain non-human-readable content.
                    #
                    # We use a threshold number of digits as a proxy for this, but
                    # don't try and get around this by just rolling other random
                    # strings; we want these outputs to be human-readable!!!
                    digit_count = len(
                        [c for c in name_in_out_dir if c in string.digits]
                    )
                    assert (
                        digit_count < 3
                    ), f"File output {name_in_out_dir} has {digit_count} digits, which is too many to be plausibly human-readable"

                assert "hello_caller.messages" in outputs
                assert outputs["hello_caller.messages"] == [
                    "Hello, Alyssa P. Hacker!",
                    "Hello, Ben Bitdiddle!",
                ]

    @needs_singularity_or_docker
    def test_miniwdl_self_test_by_reference(self, tmp_path: Path) -> None:
        """
        Test if the MiniWDL self test works when passing input files by URL reference.
        """
        self.test_miniwdl_self_test(
            tmp_path=tmp_path, extra_args=["--referenceInputs=True"]
        )

    @pytest.mark.integrative
    @needs_singularity_or_docker
    def test_dockstore_trs(
        self, tmp_path: Path, extra_args: Optional[list[str]] = None
    ) -> None:
        wdl_file = "#workflow/github.com/dockstore/bcc2020-training/HelloWorld:master"
        # Needs an input but doesn't provide a good one.
        json_input = json.dumps(
            {
                "hello_world.hello.myName": "https://raw.githubusercontent.com/dockstore/bcc2020-training/refs/heads/master/wdl-training/exercise1/name.txt"
            }
        )

        result_json = subprocess.check_output(
            self.base_command
            + [
                wdl_file,
                json_input,
                "--logDebug",
                "-o",
                str(tmp_path),
                "--outputDialect",
                "miniwdl",
            ]
            + (extra_args or [])
        )
        result = json.loads(result_json)

        with open(result.get("outputs", {}).get("hello_world.helloFile")) as f:
            result_text = f.read().strip()

        assert result_text == "Hello World!\nMy name is potato."

    # TODO: Should this move to the TRS/Dockstore tests file?
    @pytest.mark.integrative
    @needs_singularity_or_docker
    def test_dockstore_metrics_publication(
        self, tmp_path: Path, extra_args: Optional[list[str]] = None
    ) -> None:
        wdl_file = "#workflow/github.com/dockstore/bcc2020-training/HelloWorld:master"
        # Needs an input but doesn't provide a good one.
        json_input = json.dumps(
            {
                "hello_world.hello.myName": "https://raw.githubusercontent.com/dockstore/bcc2020-training/refs/heads/master/wdl-training/exercise1/name.txt"
            }
        )

        env = dict(os.environ)
        # Set credentials we got permission to publish from the Dockstore team,
        # and work on the staging Dockstore.
        env["TOIL_TRS_ROOT"] = "https://staging.dockstore.org"
        env["TOIL_DOCKSTORE_TOKEN"] = "99cf5578ebe94b194d7864630a86258fa3d6cedcc17d757b5dd49e64ee3b68c3"
        # Enable history for when <https://github.com/DataBiosphere/toil/pull/5258> merges
        env["TOIL_HISTORY"] = "True"

        output_log = subprocess.check_output(
            self.base_command
            + [
                wdl_file,
                json_input,
                "--logDebug",
                "-o",
                str(tmp_path),
                "--outputDialect",
                "miniwdl",
                "--publishWorkflowMetrics=current",
            ]
            + (extra_args or []),
            stderr=subprocess.STDOUT,
            env=env,
        )

        assert b'Workflow metrics were accepted by Dockstore.' in output_log

    @slow
    @needs_docker_cuda
    def test_giraffe_deepvariant(self, tmp_path: Path) -> None:
        """Test if Giraffe and GPU DeepVariant run. This could take 25 minutes."""
        # TODO: enable test if nvidia-container-runtime and Singularity are installed but Docker isn't.

        json_dir = tmp_path / "json"
        json_dir.mkdir()
        base_uri = "https://raw.githubusercontent.com/vgteam/vg_wdl/65dd739aae765f5c4dedd14f2e42d5a263f9267a"

        wdl_file = f"{base_uri}/workflows/giraffe_and_deepvariant.wdl"
        json_file = json_dir / "inputs.json"
        with json_file.open("w") as fp:
            # Write some inputs. We need to override the example inputs to use a GPU container, but that means we need absolute input URLs.
            json.dump(
                {
                    "GiraffeDeepVariant.INPUT_READ_FILE_1": f"{base_uri}/tests/small_sim_graph/reads_1.fastq.gz",
                    "GiraffeDeepVariant.INPUT_READ_FILE_2": f"{base_uri}/tests/small_sim_graph/reads_2.fastq.gz",
                    "GiraffeDeepVariant.XG_FILE": f"{base_uri}/tests/small_sim_graph/graph.xg",
                    "GiraffeDeepVariant.SAMPLE_NAME": "s0",
                    "GiraffeDeepVariant.GBWT_FILE": f"{base_uri}/tests/small_sim_graph/graph.gbwt",
                    "GiraffeDeepVariant.GGBWT_FILE": f"{base_uri}/tests/small_sim_graph/graph.gg",
                    "GiraffeDeepVariant.MIN_FILE": f"{base_uri}/tests/small_sim_graph/graph.min",
                    "GiraffeDeepVariant.DIST_FILE": f"{base_uri}/tests/small_sim_graph/graph.dist",
                    "GiraffeDeepVariant.OUTPUT_GAF": True,
                    "GiraffeDeepVariant.runDeepVariantCallVariants.in_dv_gpu_container": "google/deepvariant:1.3.0-gpu",
                },
                fp,
            )

        result_json = subprocess.check_output(
            self.base_command
            + [
                wdl_file,
                json_file,
                "-o",
                str(tmp_path / "out"),
                "--outputDialect",
                "miniwdl",
            ]
        )
        result = json.loads(result_json)

        # Expect MiniWDL-style output with a designated "dir"
        assert "dir" in result
        assert isinstance(result["dir"], str)
        out_dir = result["dir"]

        assert "outputs" in result
        assert isinstance(result["outputs"], dict)
        outputs = result["outputs"]

        # Expect a VCF file to have been written
        assert "GiraffeDeepVariant.output_vcf" in outputs
        assert isinstance(outputs["GiraffeDeepVariant.output_vcf"], str)
        assert os.path.exists(outputs["GiraffeDeepVariant.output_vcf"])

    @slow
    @needs_singularity_or_docker
    def test_giraffe(self, tmp_path: Path) -> None:
        """Test if Giraffe runs. This could take 12 minutes. Also we scale it down but it still demands lots of memory."""
        # TODO: enable test if nvidia-container-runtime and Singularity are installed but Docker isn't.
        # TODO: Reduce memory requests with custom/smaller inputs.
        # TODO: Skip if node lacks enough memory.

        base_uri = "https://raw.githubusercontent.com/vgteam/vg_wdl/65dd739aae765f5c4dedd14f2e42d5a263f9267a"
        wdl_file = f"{base_uri}/workflows/giraffe.wdl"
        json_file = f"{base_uri}/params/giraffe.json"

        result_json = subprocess.check_output(
            self.base_command
            + [
                wdl_file,
                json_file,
                "-o",
                str(tmp_path),
                "--outputDialect",
                "miniwdl",
                "--scale",
                "0.1",
                "--logDebug",
            ]
        )
        result = json.loads(result_json)

        # Expect MiniWDL-style output with a designated "dir"
        assert "dir" in result
        assert isinstance(result["dir"], str)
        out_dir = result["dir"]

        assert "outputs" in result
        assert isinstance(result["outputs"], dict)
        outputs = result["outputs"]

        # Expect a BAM file to have been written
        assert "Giraffe.output_bam" in outputs
        assert isinstance(outputs["Giraffe.output_bam"], str)
        assert os.path.exists(outputs["Giraffe.output_bam"])

    @needs_singularity_or_docker
    @needs_google_storage
    def test_gs_uri(self, tmp_path: Path) -> None:
        """Test if Toil can access Google Storage URIs."""
        with get_data("test/wdl/md5sum/md5sum.1.0.wdl") as wdl:
            with get_data("test/wdl/md5sum/md5sum-gs.json") as json_file:
                result_json = subprocess.check_output(
                    self.base_command
                    + [str(wdl), str(json_file), "-o", str(tmp_path), "--logDebug"]
                )
                result = json.loads(result_json)

                assert "ga4ghMd5.value" in result
                assert isinstance(result["ga4ghMd5.value"], str)
                assert os.path.exists(result["ga4ghMd5.value"])
                assert os.path.basename(result["ga4ghMd5.value"]) == "md5sum.txt"

    def test_check(self, tmp_path: Path) -> None:
        """Test that Toil's lint check works"""
        with get_data("test/wdl/lint_error.wdl") as wdl:
            out = subprocess.check_output(
                self.base_command + [str(wdl), "-o", str(tmp_path), "--logInfo"], stderr=subprocess.STDOUT)

            assert b'UnnecessaryQuantifier' in out

            p = subprocess.Popen(
                self.base_command + [wdl, "--strict=True", "--logCritical"], stderr=subprocess.PIPE)
            # Not actually a test assert; we need this to teach MyPy that we
            # get an stderr when we pass stderr=subprocess.PIPE.
            assert p.stderr is not None
            stderr = p.stderr.read()
            p.wait()
            assert p.returncode == 2
            assert b'Workflow did not pass linting in strict mode' in stderr

class TestWDLToilBench(unittest.TestCase):
    """Tests for Toil's MiniWDL-based implementation that don't run workflows."""

    def test_coalesce(self) -> None:
        """
        Test if WDLSectionJob can coalesce WDL decls.

        White box test; will need to be changed or removed if the WDL interpreter changes.
        """

        # Set up data structures for our fake workflow graph to pull from.
        # This has all decl-type nodes
        all_decls: set[str] = set()
        # And this has all transitive dependencies for all nodes.
        all_deps: dict[str, set[str]] = {}

        def mock_is_decl(self: Any, node_id: str) -> bool:
            """
            Replacement function to determine if a node is a decl or not.
            """
            return node_id in all_decls

        def mock_get_transitive_dependencies(self: Any, node_id: str) -> set[str]:
            """
            Replacement function to get all the transitive dependencies of a node.
            """
            return all_deps[node_id]

        # These are the only two methods coalesce_nodes calls, so we can
        # replace them to ensure our graph is used without actually needing to
        # make any WDL objects for it.
        #
        # If that changes, the test will need to change! Maybe then it will be
        # worth extracting a base type for this interface.
        with patch.object(WDLWorkflowGraph, "is_decl", mock_is_decl):
            with patch.object(
                WDLWorkflowGraph,
                "get_transitive_dependencies",
                mock_get_transitive_dependencies,
            ):
                with self.subTest(msg="Two unrelated decls can coalesce"):
                    # Set up two unrelated decls
                    all_decls = {"decl1", "decl2"}
                    all_deps = {"decl1": set(), "decl2": set()}

                    result = WDLSectionJob.coalesce_nodes(
                        ["decl1", "decl2"], WDLWorkflowGraph([])
                    )

                    # Make sure they coalesced
                    assert len(result) == 1
                    assert "decl1" in result[0]
                    assert "decl2" in result[0]

                with self.subTest(msg="A decl will not coalesce with a non-decl"):
                    all_decls = {"decl"}
                    all_deps = {"decl": set(), "nondecl": set()}

                    result = WDLSectionJob.coalesce_nodes(
                        ["decl", "nondecl"], WDLWorkflowGraph([])
                    )

                    assert len(result) == 2
                    assert len(result[0]) == 1
                    assert len(result[1]) == 1

                with self.subTest(
                    msg="Two adjacent decls with a common dependency can coalesce"
                ):
                    all_decls = {"decl1", "decl2"}
                    all_deps = {"decl1": {"base"}, "decl2": {"base"}, "base": set()}

                    result = WDLSectionJob.coalesce_nodes(
                        ["base", "decl1", "decl2"], WDLWorkflowGraph([])
                    )

                    assert len(result) == 2
                    assert "base" in result[0]
                    assert "decl1" in result[1]
                    assert "decl2" in result[1]

                with self.subTest(
                    msg="Two adjacent decls with different dependencies will not coalesce"
                ):
                    all_decls = {"decl1", "decl2"}
                    all_deps = {"decl1": {"base"}, "decl2": set(), "base": set()}

                    result = WDLSectionJob.coalesce_nodes(
                        ["base", "decl1", "decl2"], WDLWorkflowGraph([])
                    )

                    assert len(result) == 3
                    assert "base" in result[0]

                with self.subTest(
                    msg="Two adjacent decls with different successors will coalesce"
                ):
                    all_decls = {"decl1", "decl2"}
                    all_deps = {"decl1": set(), "decl2": set(), "successor": {"decl2"}}

                    result = WDLSectionJob.coalesce_nodes(
                        ["decl1", "decl2", "successor"], WDLWorkflowGraph([])
                    )

                    assert len(result) == 2
                    assert "decl1" in result[0]
                    assert "decl2" in result[0]
                    assert "successor" in result[1]

    def make_string_expr(self, to_parse: str) -> WDL.Expr.String:
        """
        Parse pseudo-WDL for testing whitespace removal.
        """

        pos = WDL.Error.SourcePosition("nowhere", "nowhere", 0, 0, 0, 0)

        parts: list[Union[str, WDL.Expr.Placeholder]] = re.split("(~{[^}]*})", to_parse)
        for i in range(1, len(parts), 2):
            parts[i] = WDL.Expr.Placeholder(pos, {}, WDL.Expr.Null(pos))

        return WDL.Expr.String(pos, parts)

    def test_remove_common_leading_whitespace(self) -> None:
        """
        Make sure leading whitespace removal works properly.
        """

        # For a single line, we remove its leading whitespace
        expr = self.make_string_expr(" a ~{b} c")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "a "
        assert trimmed.parts[2] == " c"

        # Whitespace removed isn't affected by totally blank lines
        expr = self.make_string_expr("    \n\n    a\n    ~{stuff}\n    b\n\n")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "\n\na\n"
        assert trimmed.parts[2] == "\nb\n\n"

        # Unless blank toleration is off
        expr = self.make_string_expr("    \n\n    a\n    ~{stuff}\n    b\n\n")
        trimmed = remove_common_leading_whitespace(expr, tolerate_blanks=False)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "    \n\n    a\n    "
        assert trimmed.parts[2] == "\n    b\n\n"

        # Whitespace is still removed if the first line doesn't have it before the newline
        expr = self.make_string_expr("\n    a\n    ~{stuff}\n    b\n")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "\na\n"
        assert trimmed.parts[2] == "\nb\n"

        # Whitespace is not removed if actual content is dedented
        expr = self.make_string_expr("    \n\n    a\n    ~{stuff}\nuhoh\n    b\n\n")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "    \n\n    a\n    "
        assert trimmed.parts[2] == "\nuhoh\n    b\n\n"

        # Unless dedents are tolerated
        expr = self.make_string_expr("    \n\n    a\n    ~{stuff}\nuhoh\n    b\n\n")
        trimmed = remove_common_leading_whitespace(expr, tolerate_dedents=True)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "\n\na\n"
        assert trimmed.parts[2] == "\nuhoh\nb\n\n"

        # Whitespace is still removed if all-whitespace lines have less of it
        expr = self.make_string_expr("\n    a\n    ~{stuff}\n  \n    b\n")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "\na\n"
        assert trimmed.parts[2] == "\n\nb\n"

        # Unless all-whitespace lines are not tolerated
        expr = self.make_string_expr("\n    a\n    ~{stuff}\n  \n    b\n")
        trimmed = remove_common_leading_whitespace(expr, tolerate_all_whitespace=False)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "\n  a\n  "
        assert trimmed.parts[2] == "\n\n  b\n"

        # When mixed tabs and spaces are detected, nothing is changed.
        expr = self.make_string_expr("\n    a\n\t~{stuff}\n    b\n")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "\n    a\n\t"
        assert trimmed.parts[2] == "\n    b\n"

        # When mixed tabs and spaces are not in the prefix, whitespace is removed.
        expr = self.make_string_expr("\n\ta\n\t~{stuff} \n\tb\n")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == "\na\n"
        assert trimmed.parts[2] == " \nb\n"

        # An empty string works
        expr = self.make_string_expr("")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 1
        assert trimmed.parts[0] == ""

        # A string of only whitespace is preserved as an all-whitespece line
        expr = self.make_string_expr("\t\t\t")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 1
        assert trimmed.parts[0] == "\t\t\t"

        # A string of only whitespace is trimmed when all-whitespace lines are not tolerated
        expr = self.make_string_expr("\t\t\t")
        trimmed = remove_common_leading_whitespace(expr, tolerate_all_whitespace=False)
        assert len(trimmed.parts) == 1
        assert trimmed.parts[0] == ""

        # An empty expression works
        expr = WDL.Expr.String(
            WDL.Error.SourcePosition("nowhere", "nowhere", 0, 0, 0, 0), []
        )
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 0

        # An expression of only placeholders works
        expr = self.make_string_expr("~{AAA}")
        trimmed = remove_common_leading_whitespace(expr)
        assert len(trimmed.parts) == 3
        assert trimmed.parts[0] == ""
        assert trimmed.parts[2] == ""

        # The command flag is preserved
        expr = self.make_string_expr(" a ~{b} c")
        trimmed = remove_common_leading_whitespace(expr)
        assert trimmed.command == False
        expr.command = True
        trimmed = remove_common_leading_whitespace(expr)
        assert trimmed.command == True

    def test_choose_human_readable_directory(self) -> None:
        """
        Test to make sure that we pick sensible but non-colliding directories to put files in.
        """

        from toil.wdl.wdltoil import (
            DirectoryNamingStateDict,
            choose_human_readable_directory,
        )

        state: DirectoryNamingStateDict = {}

        # The first time we should get  apath with the task name and without the ID
        first_chosen = choose_human_readable_directory(
            "root", "taskname", "111-222-333", state
        )
        assert first_chosen.startswith("root")
        assert "taskname" in first_chosen
        assert "111-222-333" not in first_chosen

        # If we use the same ID we should get the same result
        same_id = choose_human_readable_directory(
            "root", "taskname", "111-222-333", state
        )
        assert same_id == first_chosen

        # If we use a different ID we should get a different result still obeying the constraints
        diff_id = choose_human_readable_directory(
            "root", "taskname", "222-333-444", state
        )
        assert diff_id != first_chosen
        assert diff_id.startswith("root")
        assert "taskname" in diff_id
        assert "222-333-444" not in diff_id

    def test_uri_packing(self) -> None:
        """
        Test to make sure Toil URI packing brings through the required information.
        """

        from toil.wdl.wdltoil import pack_toil_uri, unpack_toil_uri

        # Set up a file
        file_id = FileID("fileXYZ", 123, True)
        task_path = "the_wf.the_task"
        dir_id = uuid4()
        file_basename = "thefile.txt"

        # Pack and unpack it
        uri = pack_toil_uri(file_id, task_path, dir_id, file_basename)
        unpacked = unpack_toil_uri(uri)

        # Make sure we got what we put in
        assert unpacked[0] == file_id
        assert unpacked[0].size == file_id.size
        assert unpacked[0].executable == file_id.executable

        assert unpacked[1] == task_path

        # TODO: We don't make the UUIDs back into UUID objects
        assert unpacked[2] == str(dir_id)

        assert unpacked[3] == file_basename

    def test_disk_parse(self) -> None:
        """
        Test to make sure the disk parsing is correct
        """
        # Test cromwell compatibility
        spec = "local-disk 5 SSD"
        specified_mount_point, part_size, part_suffix = parse_disks(spec, spec)
        assert specified_mount_point is None
        assert part_size == 5
        assert part_suffix == "GB"

        # Test spec conformance
        # https://github.com/openwdl/wdl/blob/e43e042104b728df1f1ad6e6145945d2b32331a6/SPEC.md?plain=1#L5072-L5082
        spec = "10"
        specified_mount_point, part_size, part_suffix = parse_disks(spec, spec)
        assert specified_mount_point is None
        assert part_size == 10
        assert part_suffix == "GiB"  # WDL spec default

        spec = "1 MB"
        specified_mount_point, part_size, part_suffix = parse_disks(spec, spec)
        assert specified_mount_point is None
        assert part_size == 1
        assert part_suffix == "MB"

        spec = "MOUNT_POINT 3"
        specified_mount_point, part_size, part_suffix = parse_disks(spec, spec)
        assert specified_mount_point == "MOUNT_POINT"
        assert part_size == 3
        assert part_suffix == "GiB"

        spec = "MOUNT_POINT 2 MB"
        specified_mount_point, part_size, part_suffix = parse_disks(spec, spec)
        assert specified_mount_point == "MOUNT_POINT"
        assert part_size == 2
        assert part_suffix == "MB"



import argparse
from typing import Any, List, OrderedDict, Sequence, TypeVar

__all__ = ["ArgumentParser", "YAMLConfigFileParser", "ConfigFileParser"]
_N = TypeVar("_N")

class ConfigFileParser(object):
    def get_syntax_description(self) -> Any: ...
    def parse(self, stream: Any) -> Any: ...
    def serialize(self, items: OrderedDict[Any, Any]) -> Any: ...

class YAMLConfigFileParser(ConfigFileParser):
    def get_syntax_description(self) -> str: ...
    def parse(self, stream: Any) -> OrderedDict[Any, Any]: ...
    def serialize(
        self, items: OrderedDict[Any, Any], default_flow_style: bool = ...
    ) -> Any: ...

class ArgumentParser(argparse.ArgumentParser):
    @property
    def _config_file_parser(self) -> Any: ...
    @property
    def _default_config_files(self) -> List[Any]: ...
    @property
    def _ignore_unknown_config_file_keys(self) -> Any: ...
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    # There may be a better way of type hinting this without a type: ignore, but mypy gets unhappy pretty much no matter what as the signatures for parse_args doesn't match with its superclass in argparse
    def parse_args(self, args: Sequence[str] | None = None, namespace: Namespace | None = None, config_file_contents: str | None = None, env_vars: Any = None) -> Namespace: ...  # type: ignore[override]
    def parse_known_args(self, args: Sequence[str] | None = ..., namespace: Namespace | None = ..., config_file_contents: Any = ..., env_vars: Any = ..., ignore_help_args: bool | None = ...) -> tuple[Namespace, list[str]]: ...  # type: ignore[override]
    def get_source_to_settings_dict(self) -> OrderedDict[Any, Any]: ...

Namespace = argparse.Namespace
ArgParser = ArgumentParser
SUPPRESS = argparse.SUPPRESS



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp


class LocalFileStoreJob(Job):
    def run(self, fileStore):
        # self.tempDir will always contain the name of a directory within the allocated disk space reserved for the job
        scratchDir = self.tempDir

        # Similarly create a temporary file.
        scratchFile = fileStore.getLocalTempFile()


if __name__ == "__main__":
    jobstore: str = mkdtemp("tutorial_managing")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "INFO"
    options.clean = "always"

    # Create an instance of FooJob which will have at least 2 gigabytes of storage space.
    j = LocalFileStoreJob(disk="2G")

    # Run the workflow
    with Toil(options) as toil:
        toil.start(j)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import subprocess
import tempfile
import time
from abc import abstractmethod
from inspect import getsource
from textwrap import dedent
from typing import TYPE_CHECKING, Optional
from uuid import uuid4

import pytest

from toil.provisioners import cluster_factory
from toil.provisioners.aws.awsProvisioner import AWSProvisioner
from toil.test import (
    ToilTest,
    get_data,
    integrative,
    needs_aws_ec2,
    needs_fetchable_appliance,
    needs_mesos,
    slow,
    timeLimit,
)
from toil.test.provisioners.clusterTest import AbstractClusterTest
from toil.version import exactPython

if TYPE_CHECKING:
    from mypy_boto3_ec2 import EC2Client
    from mypy_boto3_ec2.type_defs import (
        DescribeVolumesResultTypeDef,
        EbsInstanceBlockDeviceTypeDef,
        FilterTypeDef,
        InstanceBlockDeviceMappingTypeDef,
        InstanceTypeDef,
        VolumeTypeDef,
    )


log = logging.getLogger(__name__)


class AWSProvisionerBenchTest(ToilTest):
    """
    Tests for the AWS provisioner that don't actually provision anything.
    """

    # Needs to talk to EC2 for image discovery
    @needs_aws_ec2
    def test_AMI_finding(self):
        for zone in ["us-west-2a", "eu-central-1a", "sa-east-1b"]:
            provisioner = AWSProvisioner(
                "fakename", "mesos", zone, 10000, None, None, enable_fuse=False
            )
            ami = provisioner._discoverAMI()
            # Make sure we got an AMI and it looks plausible
            assert ami.startswith("ami-")

    @needs_aws_ec2
    def test_read_write_global_files(self):
        """
        Make sure the `_write_file_to_cloud()` and `_read_file_from_cloud()`
        functions of the AWS provisioner work as intended.
        """
        provisioner = AWSProvisioner(
            f"aws-provisioner-test-{uuid4()}",
            "mesos",
            "us-west-2a",
            50,
            None,
            None,
            enable_fuse=False,
        )
        key = "config/test.txt"
        contents = b"Hello, this is a test."

        try:
            url = provisioner._write_file_to_cloud(key, contents=contents)
            self.assertTrue(url.startswith("s3://"))

            self.assertEqual(contents, provisioner._read_file_from_cloud(key))
        finally:
            # the cluster was never launched, but we need to clean up the s3 bucket
            provisioner.destroyCluster()


@needs_aws_ec2
@needs_fetchable_appliance
@slow
@integrative
class AbstractAWSAutoscaleTest(AbstractClusterTest):
    def __init__(self, methodName):
        super().__init__(methodName=methodName)
        self.instanceTypes = ["m5a.large"]
        self.clusterName = "aws-provisioner-test-" + str(uuid4())
        self.numWorkers = ["2"]
        self.numSamples = 2
        self.spotBid = 0.15
        # We can't dump our user script right in /tmp or /home, because hot
        # deploy refuses to zip up those whole directories. So we make sure to
        # have a subdirectory to upload the script to.
        self.scriptDir = "/tmp/t"
        # Where should we put our virtualenv?
        self.venvDir = "/tmp/venv"
        # Where should we put our data to work on?
        # Must exist in the Toil container; the leader will try to rsync to it
        # (for the SSE key) and not create it.
        self.dataDir = "/tmp"
        # What filename should we use for our script (without path)?
        # Can be changed by derived tests.
        self.scriptName = "test_script.py"

    def script(self):
        """
        Return the full path to the user script on the leader.
        """
        return os.path.join(self.scriptDir, self.scriptName)

    def data(self, filename):
        """
        Return the full path to the data file with the given name on the leader.
        """
        return os.path.join(self.dataDir, filename)

    def rsyncUtil(self, src: str, dest: str) -> None:
        subprocess.check_call(
            [
                "toil",
                "rsync-cluster",
                "--insecure",
                "-p=aws",
                "-z",
                self.zone,
                self.clusterName,
            ]
            + [src, dest]
        )

    def getRootVolID(self) -> str:
        instances: list["InstanceTypeDef"] = self.cluster._get_nodes_in_cluster_boto3()
        instances.sort(key=lambda x: x.get("LaunchTime"))
        leader: "InstanceTypeDef" = instances[0]  # assume leader was launched first

        bdm: Optional[list["InstanceBlockDeviceMappingTypeDef"]] = leader.get(
            "BlockDeviceMappings"
        )
        assert bdm is not None
        root_block_device: Optional["EbsInstanceBlockDeviceTypeDef"] = None
        for device in bdm:
            if device["DeviceName"] == "/dev/xvda":
                root_block_device = device["Ebs"]
        assert (
            root_block_device is not None
        )  # There should be a device named "/dev/xvda"
        assert root_block_device.get("VolumeId") is not None
        return root_block_device["VolumeId"]

    @abstractmethod
    def _getScript(self):
        """Download the test script needed by the inheriting unit test class."""
        raise NotImplementedError()

    def putScript(self, content: str):
        """
        Helper method for _getScript to inject a script file at the configured script path, from text.
        """
        cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )
        leader = cluster.getLeader()

        self.sshUtil(["mkdir", "-p", self.scriptDir])

        with tempfile.NamedTemporaryFile(mode="w") as t:
            # use appliance ssh method instead of sshutil so we can specify input param
            t.write(content)
            # This works to make writes visible on non-Windows
            t.flush()
            leader.injectFile(t.name, self.script(), "toil_leader")

    @abstractmethod
    def _runScript(self, toilOptions):
        """
        Modify the provided Toil options to suit the test Toil script, then run the script with
        those arguments.

        :param toilOptions: List of Toil command line arguments. This list may need to be
               modified to suit the test script's requirements.
        """
        raise NotImplementedError()

    def _test(self, preemptibleJobs=False):
        """Does the work of the testing.  Many features' tests are thrown in here in no particular order."""
        # Make a cluster
        self.launchCluster()
        # get the leader so we know the IP address - we don't need to wait since create cluster
        # already insures the leader is running
        self.cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )
        self.leader = self.cluster.getLeader()
        self.sshUtil(["mkdir", "-p", self.scriptDir])
        self.sshUtil(["mkdir", "-p", self.dataDir])

        assert len(self.cluster._getRoleNames()) == 1
        # --never-download prevents silent upgrades to pip, wheel and setuptools
        venv_command = [
            "virtualenv",
            "--system-site-packages",
            "--python",
            exactPython,
            "--never-download",
            self.venvDir,
        ]
        self.sshUtil(venv_command)

        log.info("Set up script...")
        self._getScript()

        toilOptions = [
            self.jobStore,
            "--workDir=/var/lib/toil",
            "--clean=always",
            "--retryCount=2",
            "--logDebug",
            "--logFile=" + os.path.join(self.scriptDir, "sort.log"),
        ]

        if preemptibleJobs:
            toilOptions.extend(["--defaultPreemptible"])

        log.info("Run script...")
        self._runScript(toilOptions)

        assert len(self.cluster._getRoleNames()) == 1

        volumeID = self.getRootVolID()
        self.cluster.destroyCluster()
        boto3_ec2: "EC2Client" = self.aws.client(region=self.region, service_name="ec2")
        volume_filter: "FilterTypeDef" = {"Name": "volume-id", "Values": [volumeID]}
        volumes: Optional[list["VolumeTypeDef"]] = None
        for attempt in range(6):
            # https://github.com/BD2KGenomics/toil/issues/1567
            # retry this for up to 1 minute until the volume disappears
            volumes = boto3_ec2.describe_volumes(Filters=[volume_filter])["Volumes"]
            if len(volumes) == 0:
                # None are left, so they have been properly deleted
                break
            time.sleep(10)
        if volumes is None or len(volumes) > 0:
            self.fail("Volume with ID %s was not cleaned up properly" % volumeID)

        assert len(self.cluster._getRoleNames()) == 0


@integrative
@needs_mesos
@pytest.mark.timeout(1800)
class AWSAutoscaleTest(AbstractAWSAutoscaleTest):
    def __init__(self, name):
        super().__init__(name)
        self.clusterName = "provisioner-test-" + str(uuid4())
        self.requestedLeaderStorage = 80
        self.scriptName = "sort.py"

    def setUp(self):
        super().setUp()
        self.jobStore = f"aws:{self.awsRegion()}:autoscale-{uuid4()}"

    def _getScript(self):
        fileToSort = os.path.join(os.getcwd(), str(uuid4()))
        with open(fileToSort, "w") as f:
            # Fixme: making this file larger causes the test to hang
            f.write("01234567890123456789012345678901")
        with get_data("test/sort/sort.py") as sort_py:
            self.rsyncUtil(
                sort_py,
                ":" + self.script(),
            )
        self.rsyncUtil(fileToSort, ":" + self.data("sortFile"))
        os.unlink(fileToSort)

    def _runScript(self, toilOptions):
        toilOptions.extend(
            [
                "--provisioner=aws",
                "--batchSystem=mesos",
                "--nodeTypes=" + ",".join(self.instanceTypes),
                "--maxNodes=" + ",".join(self.numWorkers),
            ]
        )
        runCommand = [
            self.python(),
            self.script(),
            "--fileToSort=" + self.data("sortFile"),
            "--sseKey=" + self.data("sortFile"),
        ]
        runCommand.extend(toilOptions)
        self.sshUtil(runCommand)

    def launchCluster(self):
        # add arguments to test that we can specify leader storage
        self.createClusterUtil(
            args=["--leaderStorage", str(self.requestedLeaderStorage)]
        )

    def getRootVolID(self) -> str:
        """
        Adds in test to check that EBS volume is build with adequate size.
        Otherwise is functionally equivalent to parent.
        :return: volumeID
        """
        volumeID = super().getRootVolID()
        boto3_ec2: "EC2Client" = self.aws.client(region=self.region, service_name="ec2")
        volume_filter: "FilterTypeDef" = {"Name": "volume-id", "Values": [volumeID]}
        volumes: "DescribeVolumesResultTypeDef" = boto3_ec2.describe_volumes(
            Filters=[volume_filter]
        )
        root_volume: "VolumeTypeDef" = volumes["Volumes"][0]  # should be first
        # test that the leader is given adequate storage
        self.assertGreaterEqual(root_volume["Size"], self.requestedLeaderStorage)
        return volumeID

    @integrative
    @needs_aws_ec2
    def testAutoScale(self):
        self.instanceTypes = ["m5a.large"]
        self.numWorkers = ["2"]
        self._test()

    @integrative
    @needs_aws_ec2
    def testSpotAutoScale(self):
        self.instanceTypes = ["m5a.large:%f" % self.spotBid]
        self.numWorkers = ["2"]
        self._test(preemptibleJobs=True)

    @integrative
    @needs_aws_ec2
    def testSpotAutoScaleBalancingTypes(self):
        self.instanceTypes = ["m5.large/m5a.large:%f" % self.spotBid]
        self.numWorkers = ["2"]
        self._test(preemptibleJobs=True)


@integrative
@needs_mesos
@pytest.mark.timeout(2400)
class AWSStaticAutoscaleTest(AWSAutoscaleTest):
    """Runs the tests on a statically provisioned cluster with autoscaling enabled."""

    def __init__(self, name):
        super().__init__(name)
        self.requestedNodeStorage = 20

    def launchCluster(self):
        from toil.lib.ec2 import wait_instances_running

        self.createClusterUtil(
            args=[
                "--leaderStorage",
                str(self.requestedLeaderStorage),
                "--nodeTypes",
                ",".join(self.instanceTypes),
                "-w",
                ",".join(self.numWorkers),
                "--nodeStorage",
                str(self.requestedLeaderStorage),
            ]
        )

        self.cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )
        # We need to wait a little bit here because the workers might not be
        # visible to EC2 read requests immediately after the create returns,
        # which is the last thing that starting the cluster does.
        time.sleep(10)
        nodes: list["InstanceTypeDef"] = self.cluster._get_nodes_in_cluster_boto3()
        nodes.sort(key=lambda x: x.get("LaunchTime"))
        # assuming that leader is first
        workers = nodes[1:]
        # test that two worker nodes were created
        self.assertEqual(2, len(workers))
        # test that workers have expected storage size
        # just use the first worker
        worker = workers[0]
        boto3_ec2: "EC2Client" = self.aws.client(region=self.region, service_name="ec2")

        worker: "InstanceTypeDef" = next(wait_instances_running(boto3_ec2, [worker]))

        bdm: Optional[list["InstanceBlockDeviceMappingTypeDef"]] = worker.get(
            "BlockDeviceMappings"
        )
        assert bdm is not None
        root_block_device: Optional["EbsInstanceBlockDeviceTypeDef"] = None
        for device in bdm:
            if device["DeviceName"] == "/dev/xvda":
                root_block_device = device["Ebs"]
        assert root_block_device is not None
        assert (
            root_block_device.get("VolumeId") is not None
        )  # TypedDicts cannot have runtime type checks

        volume_filter: "FilterTypeDef" = {
            "Name": "volume-id",
            "Values": [root_block_device["VolumeId"]],
        }
        root_volume: "VolumeTypeDef" = boto3_ec2.describe_volumes(
            Filters=[volume_filter]
        )["Volumes"][
            0
        ]  # should be first
        self.assertGreaterEqual(root_volume.get("Size"), self.requestedNodeStorage)

    def _runScript(self, toilOptions):
        # Autoscale even though we have static nodes
        toilOptions.extend(
            [
                "--provisioner=aws",
                "--batchSystem=mesos",
                "--nodeTypes=" + ",".join(self.instanceTypes),
                "--maxNodes=" + ",".join(self.numWorkers),
            ]
        )
        runCommand = [
            self.python(),
            self.script(),
            "--fileToSort=" + self.data("sortFile"),
        ]
        runCommand.extend(toilOptions)
        self.sshUtil(runCommand)


@integrative
@pytest.mark.timeout(1200)
class AWSManagedAutoscaleTest(AWSAutoscaleTest):
    """Runs the tests on a self-scaling Kubernetes cluster."""

    def __init__(self, name):
        super().__init__(name)
        self.requestedNodeStorage = 20

    def launchCluster(self):
        self.createClusterUtil(
            args=[
                "--leaderStorage",
                str(self.requestedLeaderStorage),
                "--nodeTypes",
                ",".join(self.instanceTypes),
                "--workers",
                ",".join([f"0-{c}" for c in self.numWorkers]),
                "--nodeStorage",
                str(self.requestedLeaderStorage),
                "--clusterType",
                "kubernetes",
            ]
        )

        self.cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )

    def _runScript(self, toilOptions):
        # Don't use the provisioner, and use Kubernetes instead of Mesos
        toilOptions.extend(["--batchSystem=kubernetes"])
        runCommand = [
            self.python(),
            self.script(),
            "--fileToSort=" + self.data("sortFile"),
        ]
        runCommand.extend(toilOptions)
        self.sshUtil(runCommand)


@integrative
@needs_mesos
@pytest.mark.timeout(1200)
class AWSAutoscaleTestMultipleNodeTypes(AbstractAWSAutoscaleTest):
    def __init__(self, name):
        super().__init__(name)
        self.clusterName = "provisioner-test-" + str(uuid4())

    def setUp(self):
        super().setUp()
        self.jobStore = f"aws:{self.awsRegion()}:autoscale-{uuid4()}"

    def _getScript(self):
        sseKeyFile = os.path.join(os.getcwd(), "keyFile")
        with open(sseKeyFile, "w") as f:
            f.write("01234567890123456789012345678901")
        with get_data("test/sort/sort.py") as sort_py:
            self.rsyncUtil(
                sort_py,
                ":" + self.script(),
            )
        self.rsyncUtil(sseKeyFile, ":" + self.data("keyFile"))
        os.unlink(sseKeyFile)

    def _runScript(self, toilOptions):
        # Set memory requirements so that sort jobs can be run
        # on small instances, but merge jobs must be run on large
        # instances
        toilOptions.extend(
            [
                "--provisioner=aws",
                "--batchSystem=mesos",
                "--nodeTypes=" + ",".join(self.instanceTypes),
                "--maxNodes=" + ",".join(self.numWorkers),
            ]
        )
        runCommand = [
            self.python(),
            self.script(),
            "--fileToSort=/home/s3am/bin/asadmin",
            "--sortMemory=0.6G",
            "--mergeMemory=3.0G",
        ]
        runCommand.extend(toilOptions)
        runCommand.append("--sseKey=" + self.data("keyFile"))
        self.sshUtil(runCommand)

    @integrative
    @needs_aws_ec2
    def testAutoScale(self):
        self.instanceTypes = ["t2.small", "m5a.large"]
        self.numWorkers = ["2", "1"]
        self._test()


@integrative
@needs_mesos
@pytest.mark.timeout(1200)
class AWSRestartTest(AbstractAWSAutoscaleTest):
    """This test insures autoscaling works on a restarted Toil run."""

    def __init__(self, name):
        super().__init__(name)
        self.clusterName = "restart-test-" + str(uuid4())
        self.scriptName = "restartScript.py"

    def setUp(self):
        super().setUp()
        self.instanceTypes = ["t2.small"]
        self.numWorkers = ["1"]
        self.jobStore = f"aws:{self.awsRegion()}:restart-{uuid4()}"

    def _getScript(self):
        def restartScript():
            import os

            from configargparse import ArgumentParser

            from toil.job import Job

            def f0(job):
                if "FAIL" in os.environ:
                    raise RuntimeError("failed on purpose")

            if __name__ == "__main__":
                parser = ArgumentParser()
                Job.Runner.addToilOptions(parser)
                options = parser.parse_args()
                rootJob = Job.wrapJobFn(f0, cores=0.5, memory="50 M", disk="50 M")
                Job.Runner.startToil(rootJob, options)

        script = dedent("\n".join(getsource(restartScript).split("\n")[1:]))
        self.putScript(script)

    def _runScript(self, toilOptions):
        # Use the provisioner in the workflow
        toilOptions.extend(
            [
                "--provisioner=aws",
                "--batchSystem=mesos",
                "--nodeTypes=" + ",".join(self.instanceTypes),
                "--maxNodes=" + ",".join(self.numWorkers),
            ]
        )
        # clean = onSuccess
        disallowedOptions = ["--clean=always", "--retryCount=2"]
        newOptions = [
            option for option in toilOptions if option not in disallowedOptions
        ]
        try:
            # include a default memory - on restart the minimum memory requirement is the default, usually 2 GB
            command = [
                self.python(),
                self.script(),
                "--setEnv",
                "FAIL=true",
                "--defaultMemory=50000000",
            ]
            command.extend(newOptions)
            self.sshUtil(command)
        except subprocess.CalledProcessError:
            pass
        else:
            self.fail("Command succeeded when we expected failure")
        with timeLimit(600):
            command = [
                self.python(),
                self.script(),
                "--restart",
                "--defaultMemory=50000000",
            ]
            command.extend(toilOptions)
            self.sshUtil(command)

    def testAutoScaledCluster(self):
        self._test()


@integrative
@needs_mesos
@pytest.mark.timeout(1200)
class PreemptibleDeficitCompensationTest(AbstractAWSAutoscaleTest):
    def __init__(self, name):
        super().__init__(name)
        self.clusterName = "deficit-test-" + str(uuid4())
        self.scriptName = "userScript.py"

    def setUp(self):
        super().setUp()
        self.instanceTypes = [
            "m5a.large:0.01",
            "m5a.large",
        ]  # instance needs to be available on the spot market
        self.numWorkers = ["1", "1"]
        self.jobStore = f"aws:{self.awsRegion()}:deficit-{uuid4()}"

    def test(self):
        self._test(preemptibleJobs=True)

    def _getScript(self):
        def userScript():
            from toil.common import Toil
            from toil.job import Job

            # Because this is the only job in the pipeline and because it is preemptible,
            # there will be no non-preemptible jobs. The non-preemptible scaler will therefore
            # not request any nodes initially. And since we made it impossible for the
            # preemptible scaler to allocate any nodes (using an abnormally low spot bid),
            # we will observe a deficit of preemptible nodes that the non-preemptible scaler will
            # compensate for by spinning up non-preemptible nodes instead.
            #
            def job(job, disk="10M", cores=1, memory="10M", preemptible=True):
                pass

            if __name__ == "__main__":
                options = Job.Runner.getDefaultArgumentParser().parse_args()
                with Toil(options) as toil:
                    if toil.config.restart:
                        toil.restart()
                    else:
                        toil.start(Job.wrapJobFn(job))

        script = dedent("\n".join(getsource(userScript).split("\n")[1:]))
        self.putScript(script)

    def _runScript(self, toilOptions):
        toilOptions.extend(
            [
                "--provisioner=aws",
                "--batchSystem=mesos",
                "--nodeTypes=" + ",".join(self.instanceTypes),
                "--maxNodes=" + ",".join(self.numWorkers),
            ]
        )
        toilOptions.extend(["--preemptibleCompensation=1.0"])
        command = [self.python(), self.script()]
        command.extend(toilOptions)
        self.sshUtil(command)



# Copyright 2017 Oregon Health and Science University
#
# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import ftplib
import logging
import netrc
import os
from contextlib import closing
from typing import Optional, Any, cast, IO
from urllib.parse import urlparse
from urllib.request import urlopen

logger = logging.getLogger(__name__)


class FtpFsAccess:
    """
    FTP access with upload.

    Taken and modified from https://github.com/ohsu-comp-bio/cwl-tes/blob/03f0096f9fae8acd527687d3460a726e09190c3a/cwl_tes/ftp.py#L37-L251
    """
    # TODO: Properly support FTP over SSL

    def __init__(
        self, cache: Optional[dict[Any, ftplib.FTP]] = None
    ):
        """
        FTP object to handle FTP connections. By default, connect over FTP with TLS.

        :param cache: cache of generated FTP objects
        """
        self.cache = cache or {}
        self.netrc = None
        try:
            if "HOME" in os.environ:
                if os.path.exists(os.path.join(os.environ["HOME"], ".netrc")):
                    self.netrc = netrc.netrc(os.path.join(os.environ["HOME"], ".netrc"))
            elif os.path.exists(os.path.join(os.curdir, ".netrc")):
                self.netrc = netrc.netrc(os.path.join(os.curdir, ".netrc"))
        except netrc.NetrcParseError as err:
            logger.debug(err)

    def exists(self, fn: str) -> bool:
        """
        Check if a file/directory exists over an FTP server
        :param fn: FTP url
        :return: True or false depending on whether the object exists on the server
        """
        return self.isfile(fn) or self.isdir(fn)

    def isfile(self, fn: str) -> bool:
        """
        Check if the FTP url points to a file
        :param fn: FTP url
        :return: True if url is file, else false
        """
        ftp = self._connect(fn)
        if ftp:
            try:
                if not self.size(fn) is None:
                    return True
                else:
                    return False
            except ftplib.all_errors:
                return False
        return False

    def isdir(self, fn: str) -> bool:
        """
        Check if the FTP url points to a directory
        :param fn: FTP url
        :return: True if url is directory, else false
        """
        ftp = self._connect(fn)
        if ftp:
            try:
                cwd = ftp.pwd()
                ftp.cwd(urlparse(fn).path)
                ftp.cwd(cwd)
                return True
            except ftplib.all_errors:
                return False
        return False

    def open(self, fn: str, mode: str) -> IO[bytes]:
        """
        Open an FTP url.

        Only supports reading, no write support.
        :param fn: FTP url
        :param mode: Mode to open FTP url in
        :return:
        """
        if "r" in mode:
            host, port, user, passwd, path = self._parse_url(fn)
            handle = urlopen("ftp://{}:{}@{}:{}/{}".format(user, passwd, host, port, path))
            return cast(IO[bytes], closing(handle))
        # TODO: support write mode
        raise Exception("Write mode FTP not implemented")

    def _parse_url(
        self, url: str
    ) -> tuple[str, int, Optional[str], Optional[str], str]:
        """
        Parse an FTP url into hostname, username, password, and path
        :param url:
        :return: hostname, username, password, path
        """
        parse = urlparse(url)
        user = parse.username
        passwd = parse.password
        host = parse.hostname
        port = parse.port
        path = parse.path
        if host is None:
            # The URL we connect to must have a host
            raise RuntimeError(f"FTP URL does not contain a host: {url}")
        # default port is 21
        if port is None:
            port = 21
        if parse.scheme == "ftp":
            if not user and self.netrc:
                if host is not None:
                    creds = self.netrc.authenticators(host)
                    if creds:
                        user, _, passwd = creds
        if not user:
            if host is not None:
                user, passwd = self._recall_credentials(host)
                if passwd is None:
                    passwd = "anonymous@"
                    if user is None:
                        user = "anonymous"
        return host, port, user, passwd, path

    def _connect(self, url: str) -> Optional[ftplib.FTP]:
        """
        Connect to an FTP server. Handles authentication.
        :param url: FTP url
        :return: FTP object
        """
        parse = urlparse(url)
        if parse.scheme == "ftp":
            host, port, user, passwd, _ = self._parse_url(url)
            if host is None:
                # there has to be a host
                return None
            if (host, user, passwd) in self.cache:
                if self.cache[(host, user, passwd)].pwd():
                    return self.cache[(host, user, passwd)]
            ftp = ftplib.FTP_TLS()
            # Note: the FTP lib logger handles logging itself and doesn't go through our logging implementation
            ftp.set_debuglevel(1 if logger.isEnabledFor(logging.DEBUG) else 0)
            ftp.connect(host, port)
            env_user = os.getenv("TOIL_FTP_USER")
            env_passwd = os.getenv("TOIL_FTP_PASSWORD")
            if env_user:
                user = env_user
            if env_passwd:
                passwd = env_passwd
            ftp.login(user or "", passwd or "", secure=False)
            self.cache[(host, user, passwd)] = ftp
            return ftp
        return None

    def _recall_credentials(
        self, desired_host: str
    ) -> tuple[Optional[str], Optional[str]]:
        """
        Grab the cached credentials
        :param desired_host: FTP hostname
        :return: username, password
        """
        for host, user, passwd in self.cache:
            if desired_host == host:
                return user, passwd
        return None, None

    def size(self, fn: str) -> Optional[int]:
        """
        Get the size of an FTP object
        :param fn: FTP url
        :return: Size of object
        """
        ftp = self._connect(fn)
        if ftp:
            host, port, user, passwd, path = self._parse_url(fn)
            try:
                return ftp.size(path)
            except ftplib.all_errors as e:
                if str(e) == "550 SIZE not allowed in ASCII mode":
                    # some servers don't allow grabbing size in ascii mode
                    # https://stackoverflow.com/questions/22090001/get-folder-size-using-ftplib/22093848#22093848
                    ftp.voidcmd("TYPE I")
                    return ftp.size(path)
                handle = urlopen("ftp://{}:{}@{}:{}/{}".format(user, passwd, host, port, path))
                info = handle.info()
                handle.close()
                if "Content-length" in info:
                    return int(info["Content-length"])
                return None

        return None



from typing import Any

class Base:
    def as_dict(self, drop_empty: bool = ...) -> Dict[str, Any]: ...
    def as_json(self, drop_empty: bool = ...) -> str: ...
    def __init__(self) -> None: ...

class Input(Base):
    url: Optional[str] = ...
    path: Optional[str] = ...
    type: str = ...
    name: Optional[str] = ...
    description: Optional[str] = ...
    content: Any = ...
    def __init__(
        self,
        url: Optional[str],
        path: Optional[str],
        type: Optional[str],
        name: Optional[str],
        description: Optional[str],
        content: Any,
    ) -> None: ...

class Output(Base):
    url: Optional[str] = ...
    path: Optional[str] = ...
    type: str = ...
    name: Optional[str] = ...
    description: Optional[str] = ...
    def __init__(
        self,
        url: Optional[str],
        path: Optional[str],
        type: Optional[str],
        name: Optional[str],
        description: Optional[str],
    ) -> None: ...

class Resources(Base):
    cpu_cores: int = ...
    ram_gb: Union[float, int, None] = ...
    disk_gb: Union[float, int, None] = ...
    preemptible: Optional[bool] = ...
    zones: Optional[List[str]] = ...
    def __init__(
        self,
        cpu_cores: int,
        ram_gb: Union[float, int, None],
        disk_gb: Union[float, int, None],
        preemptible: Optional[bool],
        zones: Optional[List[str]],
    ) -> None: ...

class Executor(Base):
    image: str = ...
    command: str = ...
    workdir: Optional[str] = ...
    stdin: Optional[str] = ...
    stdout: Optional[str] = ...
    stderr: Optional[str] = ...
    env: Optional[Dict[str, str]] = ...
    def __init__(
        self,
        image: str,
        command: str,
        workdir: Optional[str],
        stdin: Optional[str],
        stdout: Optional[str],
        stderr: Optional[str],
        env: Optional[Dict[str, str]],
    ) -> None: ...

class ExecutorLog(Base):
    start_time: Optional[datetime] = ...
    end_time: Optional[datetime] = ...
    stdout: Optional[str] = ...
    stderr: Optional[str] = ...
    exit_code: Optional[int] = ...
    def __init__(
        self,
        start_time: Optional[datetime],
        end_time: Optional[datetime],
        stdout: Optional[str],
        stderr: Optional[str],
        exit_code: Optional[int],
    ) -> None: ...

class OutputFileLog(Base):
    url: Optional[str] = ...
    path: Optional[str] = ...
    size_bytes: Optional[int] = ...
    def __init__(
        self, url: Optional[str], path: Optional[str], size_bytes: Optional[int]
    ) -> None: ...

class TaskLog(Base):
    start_time: Optional[datetime] = ...
    end_time: Optional[datetime] = ...
    metadata: Optional[Dict[str, Any]] = ...
    logs: Optional[List[ExecutorLog]] = ...
    outputs: Optional[List[OutputFileLog]] = ...
    system_logs: Optional[List[str]] = ...
    def __init__(
        self,
        start_time: Optional[datetime],
        end_time: Optional[datetime],
        metadata: Optional[Dict[str, Any]],
        logs: Optional[List[ExecutorLog]],
        outputs: Optional[List[OutputFileLog]],
        system_logs: Optional[List[str]],
    ) -> None: ...

class Task(Base):
    id: Optional[str] = ...
    state: Optional[str] = ...
    name: Optional[str] = ...
    description: Optional[str] = ...
    inputs: Optional[List[Input]] = ...
    outputs: Optional[List[Output]] = ...
    resources: Optional[Resources] = ...
    executors: Optional[List[Executor]] = ...
    volumes: Optional[List[str]] = ...
    tags: Optional[List[str]] = ...
    logs: Optional[List[TaskLog]] = ...
    creation_time: Optional[datetime] = ...
    def is_valid(self) -> Tuple[bool, Optional[BaseException]]: ...
    def __init__(
        self,
        id: Optional[str],
        state: Optional[str],
        name: Optional[str],
        description: Optional[str],
        inputs: Optional[List[Input]],
        outputs: Optional[List[Output]],
        resources: Optional[Resources],
        executors: Optional[List[Executor]],
        volumes: Optional[List[str]],
        tags: Optional[List[str]],
        logs: Optional[List[TaskLog]],
        creation_time: Optional[datetime],
    ) -> None: ...

class GetTaskRequest(Base):
    id: Optional[str] = ...
    view: Optional[str] = ...
    def __init__(self, id: Optional[str], view: Optional[str]) -> None: ...

class CreateTaskResponse(Base):
    id: Optional[str] = ...
    def __init__(self, id: Optional[str]) -> None: ...

class ServiceInfoRequest(Base):
    def __init__(self) -> None: ...

class ServiceInfo(Base):
    name: Optional[str] = ...
    doc: Optional[str] = ...
    storage: Optional[List[str]] = ...
    def __init__(
        self, name: Optional[str], doc: Optional[str], storage: Optional[List[str]]
    ) -> None: ...

class CancelTaskRequest(Base):
    id: Optional[str] = ...
    def __init__(self, id: Optional[str]) -> None: ...

class CancelTaskResponse(Base):
    def __init__(self) -> None: ...

class ListTasksRequest(Base):
    project: Optional[str] = ...
    name_prefix: Optional[str] = ...
    page_size: Optional[int] = ...
    page_token: Optional[str] = ...
    view: Optional[str] = ...
    def __init__(
        self,
        project: Optional[str],
        name_prefix: Optional[str],
        page_size: Optional[int],
        page_token: Optional[str],
        view: Optional[str],
    ) -> None: ...

class ListTasksResponse(Base):
    tasks: Optional[List[Task]] = ...
    next_page_token: Optional[str] = ...
    def __init__(
        self, tasks: Optional[List[Task]], next_page_token: Optional[str]
    ) -> None: ...
